#   input.py - input routines for AllPlot
'''
    data is now a dict, mapping column names to their data.
    data always contains "N" (the index of each row).

    data can contain various types of data:
        column      an np array with shape equal to that of N
        parameter   an np array with shape == (1,); value is float or string
        _comment    an np array containing a single ASCII comment
        _markers    an np array containing a list of marker values
                    shape not the same as that of N (much shorter)

    Note that all items in data are np arrays.
'''

import sys, time, re, os, math
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox
from tkinter import simpledialog
import tkinter.scrolledtext as tkst
import numpy as np
from copy import *

###
### class InputParameters
### Every instance must implement three routines:
###     isValid()           returns True if this Input can be used (if 
###                         False, it does not appear in the openInput()
###                         dialog, and the other routines are never called)
###     addTab(tabs)        appends a tab into tabs, which is the 
###                         ttk.Notebook in the input dialog that permits
###                         selecting an input type. The tab widget should
###                         contain widgets that define all necessary
###                         parameter values for the Input
###     getInput()          returns a valid Input, or None
### The expectation is that XXInputParameters is the parameter object
### for the corresponding input, and also has references to the widgets of
### its tab in the openInput() dialog.
###
class InputParameters:
    def isValid(self):
        return False
    def addTab(self,tabs):
        pass
    def getInput(self):
        return None


###
### class Input - base class for all inputs
###
class Input:
    def __init__(self):
        self.filename = ""
        self.shortFilename = ""
        self.data = None
        self.selection = ""
        self.inp = None
        self.sampleRate = 0      # nonzero forces generateTimeColumn()
        self.frequencyFactor = 0 # nonzero multiplies "Frequency" column
        self.minEvent = 0
        self.maxEvent = 10000000
        self.selectionPlotColumnName = None
        self.window = None # set in AllPlotWindow.__init__
        self.fitPanel = None
        self.fit = None
    def windowClosed(self):
        if self.window is not None:
            self.window.windowClosed()
    def setWindow(self,window):
        self.window = window
        if hasattr(self,"popup") and self.popup is not None:
            window.wait_visibility(window)
            self.popup.destroy()
            self.popup = None
    def isValid(self):
        return self.inp is not None
    def hasData(self):
        return self.data is not None
    def reload(self):
        save = self.saveAllParams()
        self.read()
        self.restoreAllParams(save)
    def isChanged(self):
        return False
    def columnNames(self):      # return list of column names
        if self.data is None: return []
        N = self.data["N"]
        return [i for i in self.data.keys() if self.data[i].shape == N.shape]
    def paramNames(self):   # return list of parameter names
        if self.data is None: return []
        return [i for i in self.data.keys() if self.data[i].shape == (1,) and \
                                                not i.startswith("_")]
    def read(self):             # read the data into the data dict
        self.data = []
    def getData(self):
        data = self.getAllData()
        if data is None: return None
        sel = self.selection
        if sel == "": return data
        indexArray = eval(sel,{},data)
        ret = {}
        for i in data:
            if i.startswith("_") or len(data[i]) == 1:
                ret[i] = data[i]
            else:
                try: ret[i] = data[i][indexArray]
                except: ret[i] = data[i]
        return ret
    def nData(self):
        if self.data is None: return 0
        else: return len(self.data["N"])
    def getAllData(self):
        if self.data is None: return None
        return self.data
    def setEventLimits(self,min,max):
        self.minEvent = min
        self.maxEvent = max
    def setSampleRate(self,sampleRate):
        self.sampleRate = sampleRate
    def setFrequencyFactor(self,frequencyFactor):
        self.frequencyFactor = frequencyFactor
    # Param values are float or str
    def setParam(self,name,value):
        try:
            value = float(value)
            dt = float
        except:
            try:
                value = str(value)
                dt = object
            except:
                return
        self.data[name] = np.array([value],dtype=dt)
    # if name is not a parameter, deflt will be raised or returned
    def getParam(self,name,deflt=None):
        if name in self.data and self.data[name].shape == (1,):
                return self.data[name][0]
        if isinstance(deflt,Exception):
            raise deflt
        return deflt
    def setComment(self,comment):
        self.data["_comment"] = np.array([comment],dtype=object)
    def getComment(self):
        if "_comment" in self.data and self.data["_comment"].shape == (1,):
            return self.data["_comment"][0]
        return None
    # The selection is a string containing an experssion that selects the
    # desired data. It is an np.array expression; to select the first 1000
    #  rows: "(N > 0) & (N < 1000)"
    def setSelection(self,selection):
        self.selection = selection
        self.window.selection.config(text=selection)
    def clearSelection(self):
        self.selection = ""
        self.window.selection.config(text="")
    def setSelectionPlotColumnName(self,col):
        self.selectionPlotColumnName = col
    def getSelectionPlotColumnName(self):
        return self.selectionPlotColumnName
    def generateTimeColumn(self, sampleRate):
        self.sampleRate = sampleRate
        if not self.hasData(): self.read()
        self.data["Time"] = self.data["N"] / float(self.sampleRate)
        try: self.colNames.remove("Time")
        except: pass
        self.colNames.append("Time")
        # add a parameter
        self.setParam("SampleRate",self.sampleRate)
    def saveAllParams(self):
        save = {}
        if self.data is None:
            return save
        for k,v in self.data.items():
            if v.shape == (1,):
                save[k] = v
        return save
    def restoreAllParams(self,save):
        for k,v in save.items():
            self.data[k] = v
    def toAPIString(self):
        return "# Internal Error - toAPIString() not implemented"

###
### class DummyInput - an Input with no data
###
class DummyInput(Input):
    def __init__(self):
        Input.__init__(self)
        pass
    def toAPIString(self):
        return "\ninp = ap.DummyInput()\n"

###
### class Generate - an Input to generate a data file
###
class Generate(Input):
    class Item:
        def __init__(self,generate):
            self.generate = generate
        def insertGui(self,window):
            pass
        def addItemToGenerateList(self):
            pass
        def addToData(self,time,data):
            pass
        def toAPIString(self):
            return ""
    class DCOffset(Item):
        def __init__(self,generate):
            Generate.Item.__init__(self,generate)
        def insertGui(self,window):
            f = tk.Frame(window)
            tk.Label(f,text="D.C. Offset   Value:").pack(side=tk.LEFT)
            self.valueEntry = tk.Entry(f, width=10)
            self.valueEntry.pack(side=tk.LEFT)
            tk.Button(f,text="Add",command=self.addItemToGenerateList). \
                                                        pack(side=tk.LEFT)
            f.pack(side=tk.TOP)
        def addItemToGenerateList(self):
            try:
                gen = Generate.DCOffset(self.generate)
                gen.value = float(self.valueEntry.get())
                self.generate.addItem(gen)
            except:
                messagebox.showerror("Input Error","Invalid float value")
        def addToData(self,time,data):
            data += self.value
        def toString(self):
            return "D.C. Offset Value=%g" % self.value
        def toAPIString(self):
            return "inp.addDCOffset(%g)\n" % self.value
    class Sinewave(Item):
        def __init__(self,generate):
            Generate.Item.__init__(self,generate)
        def insertGui(self,window):
            f = tk.Frame(window)
            tk.Label(f,text="Sinewave   Amplitude:").pack(side=tk.LEFT)
            self.amplitudeEntry = tk.Entry(f, width=10)
            self.amplitudeEntry.pack(side=tk.LEFT)
            tk.Label(f,text="    Frequency(Hz):").pack(side=tk.LEFT)
            self.frequencyEntry = tk.Entry(f,width=10)
            self.frequencyEntry.pack(side=tk.LEFT)
            tk.Label(f,text="    Phase(deg):").pack(side=tk.LEFT)
            self.phaseEntry = tk.Entry(f,width=10)
            self.phaseEntry.pack(side=tk.LEFT)
            tk.Button(f,text="Add",command=self.addItemToGenerateList). \
                                                        pack(side=tk.LEFT)
            f.pack(side=tk.TOP)
        def addItemToGenerateList(self):
            try:
                gen = Generate.Sinewave(self.generate)
                gen.amplitude = float(self.amplitudeEntry.get())
                gen.frequency = float(self.frequencyEntry.get())
                gen.phase = float(self.phaseEntry.get()) * math.pi / 180.0
                self.generate.addItem(gen)
            except:
                messagebox.showerror("Input Error","Invalid float value")
        def addToData(self,time,data):
            w = self.frequency * 2.0 * math.pi
            data += self.amplitude * np.sin(w*time + self.phase)
        def toString(self):
            return "Sinewave Amplitude=%g Frequency=%g Phase=%g" % \
                                (self.amplitude,self.frequency,self.phase)
        def toAPIString(self):
            return "inp.addSinewave(%g,%g,%g)\n" % \
                        (self.amplitude,self.frequency,self.phase)
    class WhiteNoise(Item):
        def __init__(self,generate):
            Generate.Item.__init__(self,generate)
        def insertGui(self,window):
            f = tk.Frame(window)
            tk.Label(f,text="White Noise   Amplitude:").pack(side=tk.LEFT)
            self.amplitudeEntry = tk.Entry(f, width=10)
            self.amplitudeEntry.pack(side=tk.LEFT)
            tk.Button(f,text="Add",command=self.addItemToGenerateList). \
                                                        pack(side=tk.LEFT)
            f.pack(side=tk.TOP)
        def addItemToGenerateList(self):
            try:
                gen = Generate.WhiteNoise(self.generate)
                gen.amplitude = float(self.amplitudeEntry.get())
                self.generate.addItem(gen)
            except:
                messagebox.showerror("Input Error","Invalid float value")
        def addToData(self,time,data):
            data += np.random.uniform(-self.amplitude,self.amplitude,len(data))
        def toString(self):
            return "White Noise Amplitude=%g" % self.amplitude
        def toAPIString(self):
            return "inp.addWhiteNoise(%g)\n" % self.amplitude
    class GaussianNoise(Item):
        def __init__(self,generate):
            Generate.Item.__init__(self,generate)
        def insertGui(self,window):
            f = tk.Frame(window)
            tk.Label(f,text="Gaussian Noise   Mean:").pack(side=tk.LEFT)
            self.meanEntry = tk.Entry(f, width=10)
            self.meanEntry.pack(side=tk.LEFT)
            tk.Label(f,text="      Sigma:").pack(side=tk.LEFT)
            self.sigmaEntry = tk.Entry(f, width=10)
            self.sigmaEntry.pack(side=tk.LEFT)
            tk.Button(f,text="Add",command=self.addItemToGenerateList). \
                                                        pack(side=tk.LEFT)
            f.pack(side=tk.TOP)
        def addItemToGenerateList(self):
            try:
                gen = Generate.GaussianNoise(self.generate)
                gen.mean = float(self.meanEntry.get())
                gen.sigma = float(self.sigmaEntry.get())
                self.generate.addItem(gen)
            except:
                messagebox.showerror("Input Error","Invalid float value")
        def addToData(self,time,data):
            data += np.random.normal(self.mean,self.sigma,len(data))
        def toString(self):
            return "Gaussian Noise Mean=%g Sigma=%g" % (self.mean,self.sigma)
        def toAPIString(self):
            return "inp.addGaussianNoise(%g,%g)\n" % (self.mean,self.sigma)
    # class Generate
    def __init__(self):
        Input.__init__(self)
        self.filename = "Generate"
        self.shortFilename = "Generate"
        self.itemList = []
        self.setSelectionPlotColumnName("Data")
        self.nSamples = 100000
        self.sampleRate = 1000
        self.nSamplesEntry = None
        self.sampleRateEntry = None
    def setWindow(self,window):
        global GenerateInstructions
        Input.setWindow(self,window)
        self.window.plotList.append(None)   # make plot indices match up
        self.frame = tk.Frame(self.window.tabs)
        self.window.tabs.add(self.frame,text="Generate")
        self.window.tabs.select(self.frame)
        # fill self.frame with GUI to generate data
        tk.Label(self.frame,text=GenerateInstructions).pack()
        f = tk.Frame(self.frame)
        tk.Label(f,text="Sample Rate (Hz):").pack(side=tk.LEFT)
        self.sampleRateEntry = tk.Entry(f,width=6)
        self.sampleRateEntry.pack(side=tk.LEFT)
        self.sampleRateEntry.insert(tk.END,"1000")
        tk.Label(f,text="      # Samples:").pack(side=tk.LEFT)
        self.nSamplesEntry = tk.Entry(f,width=10)
        self.nSamplesEntry.pack(side=tk.LEFT)
        self.nSamplesEntry.insert(tk.END,"100000")
        f.pack(side=tk.TOP)
        tk.Label(self.frame,text=" ").pack(side=tk.TOP)
        tk.Button(self.frame,text="Clear All Items",command=self.clearItems). \
                                                    pack(side=tk.TOP)
        # Items to add
        tk.Label(self.frame,text="Items to add:").pack(side=tk.TOP)
        self.dcoffset = self.DCOffset(self)
        self.dcoffset.insertGui(self.frame)
        self.sinewave = self.Sinewave(self)
        self.sinewave.insertGui(self.frame)
        self.gaussianNoise = self.GaussianNoise(self)
        self.gaussianNoise.insertGui(self.frame)
        self.whiteNoise = self.WhiteNoise(self)
        self.whiteNoise.insertGui(self.frame)
        tk.Label(self.frame,text="\nList of items to be summed:"). \
                                                        pack(side=tk.TOP)
        self.textList = tkst.ScrolledText(self.frame,width=80,height=10)
        self.textList.configure(state ='disabled')
        self.textList.pack(side=tk.TOP,fill=tk.Y,expand=True)
        for item in self.itemList:
            self.textList.configure(state ='normal')
            self.textList.insert(tk.END,item.toString()+"\n")
            self.textList.configure(state ='disabled')
        tk.Button(self.frame,text="Generate New Data and Replot All Plots",
                                command=self.doGenerate).pack(side=tk.TOP)
        self.doGenerate()
    def addItem(self,item):
        self.itemList.append(item)
        if hasattr(self,'textList'):
            self.textList.configure(state ='normal')
            self.textList.insert(tk.END,item.toString()+"\n")
            self.textList.configure(state ='disabled')
    def clearItems(self):
        self.itemList.clear()
        if hasattr(self,'textList'):
            self.textList.configure(state ='normal')
            self.textList.delete('1.0',tk.END)
            self.textList.configure(state ='disabled')
    def doGenerate(self):
        if self.nSamplesEntry is not None:
            try:
                self.nSamples = int(self.nSamplesEntry.get())
                self.sampleRate = float(self.sampleRateEntry.get())
            except:
                messagebox.showerror("Input Error","Invalid sample value")
                return
        self.data = {}
        self.data["N"] = np.arange(self.nSamples)
        self.data["Time"] = self.data["N"] / self.sampleRate
        self.data["Data"] = np.zeros(self.nSamples)
        for i in self.itemList:
            i.addToData(self.data["Time"],self.data["Data"])
        if self.window is not None:
            for p in self.window.plotList:
                if p is not None: p.reload()   # autoscale & draw()
    def setSamples(self,nSamples, sampleRate):
        self.nSamples = nSamples
        self.sampleRate = sampleRate
        self.doGenerate()
    def addDCOffset(self,value):
        item = Generate.DCOffset(self)
        item.value= value
        self.addItem(item)
    def addSinewave(self,amplitude, frequency, phase):
        item = Generate.Sinewave(self)
        item.amplitude = amplitude
        item.frequency = frequency
        item.phase = phase
        self.addItem(item)
    def addGaussianNoise(self,mean,sigma):
        item = Generate.GaussianNoise(self)
        item.mean = mean
        item.sigma = sigma
        self.addItem(item)
    def addWhiteNoise(self,amplitude):
        item = Generate.WhiteNoise(self)
        item.amplitude = amplitude
        self.addItem(item)
    def toAPIString(self):
        ret = "\n"
        ret += "inp = ap.Generate()\n"
        ret += "inp.setSamples(%g,%g)\n" % (self.nSamples,self.sampleRate)
        for i in self.itemList:
            ret += i.toAPIString()
        ret += "inp.doGenerate()\n"
        return ret

GenerateInstructions = '''
Generate will create a time-series data file consisting of the sum of the
items added below.
'''
