#   plot.py - plot routines for AllPlot
'''
    Each plot class has all the arguments needed to specify the plot. It ia
    followed by a dialog box class to select parameters; if OK is pressed
    the plot is created in the apply() function, as a new tab in the window.

    Each plot class constructor has an optional argument forceAxes --
    if None then the constructor creates a new tab in the Window and
    sets up the Frame for the plot and calls draw(); if not None, the
    constructor does not create a new tab, does not setup any Frame, and 
    calls draw(forceAces). This is used for making multiple plots in
    a single window.

    Each plot class has a method draw(axes=None) -- if None then the
    plot is drawn in self.axes (set by the constructor); if not None then
    the plot is drawn in the given axes.

    NOTE: if the parameter FrequencyScale is present, it means that the 
    Frequency column has already been multiplied by its value; the
    corresponding unit is in param FrequencyUnit. The unit is added to
    axis labels that are "Frequency" in Plot.applyUnit(). Also if params
    FrequencyScale and HzPerPicometer are present, the AllanDeviationPlot
    is converted from Frequency to picometers.

'''

import sys, time, re, os, math
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import simpledialog
import numpy as np
import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import RectangleSelector
import numpy.fft as fft

import input

#try:
#    import allantools
#    Allantools = True
#except:
#    Allantools = False
import allantools
Allantools = True

plt.ion()

PlotList = []

# np array functions usable in expressions
Functions = { "abs":np.abs, "arccos":np.arccos, "arccosh":np.arccosh,
    "arcsin":np.arcsin, "arcsinh":np.arcsinh, "arctan":np.arctan,
    "arctan2":np.arctan2, "arctanh":np.arctanh, "min":np.min,
    "max":np.max, "ceil":np.ceil, "floor":np.floor, "cos":np.cos,
    "cosh":np.cosh, "deg2rad":np.deg2rad, "degrees":np.degrees,
    "exp":np.exp, "exp2":np.exp2, "fabs":np.fabs, "log":np.log,
    "log10":np.log10, "log2":np.log2, "power":np.power,
    "rad2deg":np.rad2deg, "sin":np.sin, "sinh":np.sinh, "sinc":np.sinc,
    "sqrt":np.sqrt, "tan":np.tan, "tanh":np.tanh, "hypot":np.hypot,
    "arctan2":np.arctan2 }

###
### setupPlotMenu()
###
def ignore():
    pass
def setupPlotMenu(window):
    plotmenu = tk.Menu(window.menubar)
    window.menubar.add_cascade(label="Plot", menu=plotmenu)
    plotmenu.add_command(label="Simple Data Plot", accelerator="Ctrl-D",
                command=lambda: SimpleDataPlotDialog(window))
    window.bind_all("<Control-d>", lambda event: SimpleDataPlotDialog(window))
    plotmenu.add_command(label="Histogram", accelerator="Ctrl-H",
                command=lambda: HistogramDialog(window))
    window.bind_all("<Control-h>", lambda event: HistogramDialog(window))
    plotmenu.add_command(label="XY Plot", accelerator="Ctrl-X",
                command=lambda: XYPlotDialog(window))
    window.bind_all("<Control-x>", lambda event: XYPlotDialog(window))
    plotmenu.add_command(label="Time Series Plots:", command=ignore)
    plotmenu.entryconfig("Time Series Plots:", state="disabled")
    plotmenu.add_command(label="FFT", accelerator="Ctrl-F",
                command=lambda: FFTPlotDialog(window))
    window.bind_all("<Control-f>", lambda event: FFTPlotDialog(window))
    plotmenu.add_command(label="Correlation", accelerator="Ctrl-C",
                command=lambda: CorrelationPlotDialog(window))
    window.bind_all("<Control-c>", lambda event: CorrelationPlotDialog(window))
    if Allantools:
        plotmenu.add_command(label="Allan Deviation", accelerator="Ctrl-A",
                command=lambda: AllanDevPlotDialog(window))
        window.bind_all("<Control-a>", lambda event: AllanDevPlotDialog(window))
    plotmenu.add_separator()
    plotmenu.add_command(label="Combination Plots:", command=ignore)
    plotmenu.entryconfig("Combination Plots:", state="disabled")
    plotmenu.add_command(label="Combo Plot",
                command=lambda: ComboPlotDialog(window))
    plotmenu.add_command(label="Quad Plot", accelerator="Ctrl-Q",
                command=lambda: QuadPlotDialog(window))
    window.bind_all("<Control-q>", lambda event: QuadPlotDialog(window))
    plotmenu.add_command(label="Compare Plot",
                command=lambda: ComparePlotDialog(window))
    plotmenu.add_separator()
    plotmenu.add_command(label="Redraw Plot", command=lambda: redraw(window))
    plotmenu.add_command(label="Edit Plot Style", accelerator="Ctrl-S",
                command=lambda: window.currentPlot().editStyle())
    window.bind_all("<Control-s>", 
                lambda event: window.currentPlot().editStyle())
    plotmenu.add_separator()
    plotmenu.add_command(label="Delete Current Plot", command=lambda: deletePlot(window))


###
### redraw()
###
def redraw(window):
    i = window.tabs.index("current")
    plot = window.plotList[i]
    if plot is not None:
        plot.draw()

###
### deletePlot()
###
def deletePlot(window):
    plot = window.currentPlot()
    if plot is None: return
    if plot.fit is not None: plot.fit = None
    if plot.fitPanel is not None: plot.fitPanel.destroy()
    i = window.tabs.index("current")
    window.tabs.forget(i)
    del window.plotList[i]

###
### validateExpr()
###
def validateExpr(expr,data):
    try:
        if expr is None or expr == "": return False
        n = 3
        d = {}
        for i in data:
            try: d[i] = data[i][0:n]
            except: d[i] = data[i]  # parameter or comment
        v = eval(expr,Functions,d)
        return True
    except:
        return False

###
### errorBox()
###
def errorBox(msg):
    messagebox.showerror(title="Error",message=msg)

###
### thinPlotData()
###
def thinPlotData(x,y,width,xmin=None,xmax=None):
    ''' Thin the drawing of a plot with more than 10000 data points
        by avoiding drawing many lines within a single pixel column. 
        Use the window width for nx rather than the axes width, because
        the latter's mapping is unreliable when this function is called;
        using more columns than the actual plot avoide the potential
        moire' pattern between the columns used here and the actual
        pixel columns of the drawing.
        This is ineffective if the x array is not sorted, or if
        the plot will be drawn using pointMarkers.
        Note the Python code never loops over the data; it uses
        np.array functions and indexing for that (MUCH faster).
    '''
    # do nothing if too small or not sorted
    if len(x) < 10000: return (x,y)
    if not np.all(np.diff(x) >= 0): return (x,y)
    try:
        if xmin is None: xmin = np.min(x)
        if xmax is None: xmax = np.max(x)
        # get xdis = pixel column where this point would be displayed (int32)
        nx = width
        a = nx / (xmax - xmin)
        #print("thin a,xmin,xmax=",a,xmin,xmax)
        xdis = (a*(x-xmin)).astype(np.int32)
        # now plot lines from min to max in each pixel column
        # note we are plotting (x,y) values; xdis is used only for thinning
        xplot = []
        yplot = []
        # loop only over pixel columns within the x bounds
        i0 = max(np.min(xdis),0)
        i1 = min(np.max(xdis)+1,nx)
        #print("thin i0,i1=",i0,i1)
        for i in range(i0,i1):
            indexArray = (xdis == i)
            xthis = x[indexArray]  # array of x values for which xdis==i
            ythis = y[indexArray]  # array of y values for which xdis==i
            if len(xthis) == 0: continue
            xmin = np.min(xthis)
            xmax = np.max(xthis)
            ymin = np.min(ythis)
            ymax = np.max(ythis)
            xplot.append(xmin)
            yplot.append(ymin)
            if xmin != xmax or ymin != ymax:
                xplot.append(xmax)
                yplot.append(ymax)
    except:
        return (x,y)
    return (xplot,yplot)

###
### class Style - define the style of a plot
### Every plot class should have style as a class field.
### Style.insertEditor(frame) will insert an editor into frame (calls pack()).
###
class Style:
    colors = ["blue","red","green","black","cyan","magenta","yellow"]
    linestyles = [" ","solid","dashed","dashdot","dotted"]
    pointMarkers = {" ":" ", "dot":".", "smalldot":",", "bigdot":"o",
                        "square":"s", "plus":"+", "x":"x", "diamond":"D"}
    def __init__(self, color='blue', linestyle='solid', pointMarker=' '):
        self.color = color
        self.linestyle = linestyle
        self.pointMarker = pointMarker
        self.colorCB = None
        self.linestyleCB = None
        self.pointMarkerCB = None
    def isDefault(self):
        return self.color == "blue" and self.linestyle == "solid" and \
               self.pointMarker == " "
    def insertEditor(self,frame):
        f = tk.Frame(frame)
        tk.Label(f,text="Style").grid(row=0,column=0,columnspan=99)
        tk.Label(f,text="Color").grid(row=1,column=0)
        self.colorCB =  ttk.Combobox(f,values=self.colors,
                                                    state="readonly",width=8)
        self.colorCB.current(self.colors.index(self.color))
        self.colorCB.grid(row=1,column=1)
        tk.Label(f,text="Line style").grid(row=2,column=0)
        self.linestyleCB = ttk.Combobox(f,values=self.linestyles,
                                                    state="readonly",width=8)
        self.linestyleCB.current(self.linestyles.index(self.linestyle))
        self.linestyleCB.grid(row=2,column=1)
        tk.Label(f,text="PointMarkers").grid(row=3,column=0)
        self.pointMarkerCB = ttk.Combobox(f,values=list(self.pointMarkers.keys()),
                                                    state="readonly",width=8)
        v = list(self.pointMarkers.values())
        self.pointMarkerCB.current(v.index(self.pointMarker))
        self.pointMarkerCB.grid(row=3,column=1)
        f.pack()
    def retrieve(self):
        if self.colorCB is None: return
        self.color = self.colorCB.get()
        self.linestyle = self.linestyleCB.get()
        self.pointMarker = self.pointMarkers[self.pointMarkerCB.get()]
    def toAPIString(self):
        return "style = ap.Style('%s','%s','%s')\n" % \
                                 (self.color,self.linestyle,self.pointMarker)

###
### class AxesParam - parameters for an Axes
### Every plot class should have axisparam as a class field.
### AxesParam.insertEditor(frame) will insert an editor into frame 
### AxesParam.apply(axes,x,y) will apply the parameters to the axes --
###     x and y are used only if autoscale is on, y can be None for a histogram.
###
class AxesParam:
    def __init__(self, title="", xlabel="", ylabel="", logx=False, logy=False,
                 gridlines=True, xmin=None, xmax=None, ymin=None, ymax=None):
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.logx =tk.IntVar()
        self.logx.set(int(logx))
        self.logy = tk.IntVar()
        self.logy.set(int(logy))
        self.gridlines = tk.IntVar()
        self.gridlines.set(int(gridlines))
        self.autoscale = tk.IntVar()
        if xmin is None and xmax is None and ymin is None and ymax is None:
            self.autoscale.set(1)
            self.xmin = 0.0
            self.xmax = 1.0
            self.ymin = 0.0
            self.ymax = 1.0
        else:
            self.autoscale.set(0)
            self.xmin = xmin
            self.xmax = xmax
            self.ymin = ymin
            self.ymax = ymax
    def insertEditor(self,master):
        f = tk.Frame(master)
        tk.Label(f,text="Title:").grid(row=1,column=0)
        self.titleE = tk.Entry(f,width=30)
        self.titleE.grid(row=1,column=1,columnspan=99)
        self.titleE.insert(0,self.title)
        tk.Label(f,text="X Label:").grid(row=2,column=0)
        self.xlabelE = tk.Entry(f,width=30)
        self.xlabelE.grid(row=2,column=1,columnspan=99)
        self.xlabelE.insert(0,self.xlabel)
        tk.Label(f,text="Y Label:").grid(row=3,column=0)
        self.ylabelE = tk.Entry(f,width=30)
        self.ylabelE.grid(row=3,column=1,columnspan=99)
        self.ylabelE.insert(0,self.ylabel)
        tk.Checkbutton(f,text="Log X", variable=self.logx,
                command=self.autoscaleChanged).grid(row=6,column=0)
        tk.Checkbutton(f,text="Log Y", variable=self.logy,
                command=self.autoscaleChanged).grid(row=6,column=1)
        tk.Checkbutton(f,text="Gridlines", variable=self.gridlines,
                command=self.autoscaleChanged).grid(row=6,column=2)
        tk.Checkbutton(f,text="Auto-scale", variable=self.autoscale,
                command=self.autoscaleChanged). \
                grid(row=6,column=3,sticky=tk.W)
        ff = self.limitsFrame = tk.Frame(f)
        tk.Label(ff,text="X min:").grid(row=0,column=0)
        self.xminE = tk.Entry(ff,width=8)
        self.xminE.grid(row=0,column=1)
        self.xminE.insert(0,str(self.xmin))
        tk.Label(ff,text="X max:").grid(row=0,column=2)
        self.xmaxE = tk.Entry(ff,width=8)
        self.xmaxE.grid(row=0,column=3)
        self.xmaxE.insert(0,str(self.xmax))
        tk.Label(ff,text="Y min:").grid(row=1,column=0)
        self.yminE = tk.Entry(ff,width=8)
        self.yminE.grid(row=1,column=1)
        self.yminE.insert(0,str(self.ymin))
        tk.Label(ff,text="Y max:").grid(row=1,column=2)
        self.ymaxE = tk.Entry(ff,width=8)
        self.ymaxE.grid(row=1,column=3)
        self.ymaxE.insert(0,str(self.ymax))
        ff.grid(row=11,column=0,columnspan=99)
        self.autoscaleChanged()
        f.pack()
    def autoscaleChanged(self):
        if self.autoscale.get() != 0:
            self.limitsFrame.grid_remove()
        else:
            self.limitsFrame.grid()
    def retrieve(self):
        try: 
            if self.titleE is None: return
        except: 
            return
        self.title = self.titleE.get()
        self.xlabel = self.xlabelE.get()
        self.ylabel = self.ylabelE.get()
        self.xmin = float(self.xminE.get())
        self.xmax = float(self.xmaxE.get())
        self.ymin = float(self.yminE.get())
        self.ymax = float(self.ymaxE.get())
    def apply(self,axes,x=None,y=None):
        axes.set_title(self.title)
        self.xlabel = self.xlabel
        axes.set_xlabel(self.xlabel)
        self.ylabel = self.ylabel
        axes.set_ylabel(self.ylabel)
        try:   axes.get_xaxis().get_major_formatter().set_useOffset(False)
        except AttributeError: pass
        try:   axes.get_yaxis().get_major_formatter().set_useOffset(False)
        except AttributeError: pass
        if self.logx.get() == 0:
            axes.set_xscale('linear')
        else:
            axes.set_xscale('log')
        if self.logy.get() == 0:
            axes.set_yscale('linear')
        else:
            axes.set_yscale('log')
        if self.gridlines.get() == 0:
            axes.grid(False)
        else:
            axes.grid(True)
        axes.autoscale(enable=False)
        if self.autoscale.get() != 0:
            self.autoscale.set(0)
            if x is not None:
                x1 = np.min(x)
                x2 = np.max(x)
                if self.logx.get() == 0:
                    dx = (x2-x1)/20
                    if dx == 0.0: dx = abs(x1/20.0)
                    if dx == 0.0: dx = 0.1
                else:
                    dx = 0.0
                    x2 *= 1.2
                self.xmin = x1-dx
                self.xmax = x2+dx
                axes.set_xbound(lower=self.xmin,upper=self.xmax)
            else:
                axes.autoscale(enable=True,axis='x')
            if y is not None:
                y1 = np.min(y)
                y2 = np.max(y)
                if self.logy.get() == 0:
                    dy = (y2-y1)/20
                    if dy == 0.0: dy = abs(y1/20.0)
                    if dy == 0.0: dy = 0.1
                else:
                    dy = 0.0
                    y2 *= 1.2
                self.ymin = y1-dy
                self.ymax = y2+dy
                axes.set_ybound(lower=self.ymin,upper=self.ymax)
            else:
                axes.autoscale(enable=True,axis='y')
        else:
            axes.set_xbound(lower=self.xmin,upper=self.xmax)
            axes.set_ybound(lower=self.ymin,upper=self.ymax)
    def setLimits(self, xmin, xmax, ymin, ymax):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.autoscale.set(0)
    def toAPIString(self):
        # logx or logy needs autoscale
        if self.logx.get() != 0 or self.logy.get() != 0:
            return "axesparam = ap.AxesParam('%s','%s','%s',%d,%d,%d)\n" % \
                (self.title,self.xlabel,self.ylabel,self.logx.get(),
                 self.logy.get(),self.gridlines.get())
        return "axesparam = ap.AxesParam('%s','%s','%s',%d,%d,%d,%g,%g,%g,%g)\n" % \
                (self.title,self.xlabel,self.ylabel,self.logx.get(),
                 self.logy.get(),self.gridlines.get(),
                 self.xmin,self.xmax,self.ymin,self.ymax)


###
### class Plot - base class for all Plot-s
###
class Plot:
    def __init__(self,window,tabName="???",forceAxes=None):
        self.window = window
        self.number = window.plotNumber()
        self.zoomStack = []
        self.tabName = tabName
        self.style = Style()
        self.axesparam = AxesParam()
        self.type = "?"
        self.name = ""
        self.topMessage = None
        self.fit = None
        self.fitPanel = None
        self.markerPanel = None
        self.mouse_click_callbacks = []
        self.mouse_move_callbacks = []
        self.draw_callbacks = []
        self.markerLines = []
        self.outlierLines = []
        self.shiftKey = None
        if forceAxes is None:
            if self not in window.plotList:
                window.plotList.append(self)
            self.frame = tk.Frame(self.window.tabs)
            self.window.tabs.add(self.frame,text=self.tabName)
            self.window.tabs.select(self.frame)
            self.axes = None
        else:
            self.axes = forceAxes
            self.frame = None
    def register_click_callback(self,fun):
        self.mouse_click_callbacks.append(fun);
    def register_move_callback(self,fun):
        self.mouse_move_callbacks.append(fun);
    def register_draw_callback(self,fun):
        self.draw_callbacks.append(fun);
    def unregister_click_callback(self,fun):
        try: self.mouse_clck_callbacks.remove(fun)
        except: pass
    def unregister_move_callback(self,fun):
        try: self.mouse_move_callbacks.remove(fun)
        except: pass
    def unregister_draw_callback(self,fun):
        try: self.draw_callbacks.remove(fun)
        except: pass
    def draw(self,axes=None, label=None):
        pass
    def callDrawCallbacks(self):
        for f in self.draw_callbacks:
            f(self)
    def setStyle(self,style,redraw=True):
        self.style = style
        if redraw: self.draw()
    def drawMarkers(self):
        # use fitValues() to determine the X expression
        x,xExpr,y,yExpr,yerr,yerrExpr = self.fitValues()
        if xExpr is None or xExpr != "Time": return
        for i in self.markerLines:
            try: i.remove()
            except: pass
        self.markerLines = []
        data = self.window.input.getData()
        if "_markers" in data:
            for i in data["_markers"]:
                line = self.axes.axvline(x=i,linestyle='-',color='red')
                self.markerLines.append(line)
    def drawOutliers(self):
        # use fitValues() to determine the X and Y expressions
        x,xExpr,y,yExpr,yerr,yerrExpr = self.fitValues()
        if xExpr is None or xExpr != "Time": return
        if yExpr is None or yExpr != "Frequency": return
        for i in self.outlierLines:
            try: i.remove()
            except: pass
        self.outlierLines = []
        data = self.window.input.getData()
        # data["_outliers"] contains the Time of an upper outlier, and
        # -Time of a lower outlier
        if "_outliers" in data:
            x = []
            y = []
            dy = (self.axesparam.ymax - self.axesparam.ymin) / 40.0
            top = self.axesparam.ymax - dy
            bottom = self.axesparam.ymin + dy
            for v in data["_outliers"]:
                if v >= 0.0:
                    x.append(v)
                    y.append(top)
                else:
                    x.append(-v)
                    y.append(bottom)
            line = self.axes.plot(x, y, color='red', linestyle="", marker='D')
            self.outlierLines.append(line)
                
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def isSorted(self,x):
        return np.all(np.diff(x) >= 0)
    def thin(self,x,y):
        ''' Thin the drawing of a plot with more than 10000 data points
            by avoiding drawing many lines within a single pixel column. 
            Use the window width for nx rather than the axes width, because
            the latter's mapping is unreliable when this function is called;
            using more columns than the actual plot avoids the potential
            moire' pattern between the columns used here and the actual
            pixel columns of the drawing.
            This is ineffective if the x array is not sorted, or if
            the plot will be drawn using pointMarkers.
            Note the Python code never loops over the data; it uses
            np.array functions and indexing for that (MUCH faster).
        '''
        try:
            if not self.isSorted(x) or len(x) < 10000: return (x,y)
            if self.axesparam.logx.get() != 0: return (x,y)
            if self.axesparam.autoscale.get() != 0: return (x,y)
            if self.style.pointMarker.strip() != "": return (x,y)
            axes = self.axes
            # get xdis = pixel column where this point would be displayed (int32)
            nx = self.window.winfo_width()
            a = nx / (self.axesparam.xmax - self.axesparam.xmin)
            #print("thin a,xmin,xmax=",a,self.axesparam.xmin,self.axesparam.xmax)
            xdis = (a*(x-self.axesparam.xmin)).astype(np.int32)
            # now plot lines from min to max in each pixel column
            # note we are plotting (x,y) values; xdis is used only for thinning
            xplot = []
            yplot = []
            # loop only over pixel columns within the x bounds
            i0 = max(np.min(xdis),0)
            i1 = min(np.max(xdis)+1,nx)
            #print("thin i0,i1=",i0,i1)
            for i in range(i0,i1):
                indexArray = (xdis == i)
                xthis = x[indexArray]  # array of x values for which xdis==i
                ythis = y[indexArray]  # array of y values for which xdis==i
                #print(self.name+" i=",i,"len(xthis)=",len(xthis))
                if len(xthis) == 0: continue
                x0 = np.min(xthis)
                x1 = np.max(xthis)
                y0 = np.min(ythis)
                y1 = np.max(ythis)
                xplot.append(x0)
                yplot.append(y0)
                if x0 != x1 or y0 != y1:
                    xplot.append(x1)
                    yplot.append(y1)
        except:
            return (x,y)
        return (xplot,yplot)
    def getName(self):
        if self.name != "": return self.name
        if hasattr(self.axesparam,"title") and  self.axesparam.title != "":
            return self.axesparam.title
        return self.type
    def checkSampleRate(self,x,sampleRate):
        # average of deltas in x must be within 0.1% of 1/sampleRate,
        # and individual deltas must be within 10% of the average
        avg = float(x[-1]-x[0])/float(len(x)-1)
        delta = avg/10.0
        m = np.amax(np.absolute(np.diff(x)-avg))
        if abs(avg - 1.0/float(sampleRate)) > avg/1000.0: return False
        return m <= delta
    def applyUnit(self,label):
        if label == "Frequency":
            unit = self.window.input.getParam("FrequencyUnit",None)
            if unit is not None:
                label += " (" + unit + ")"
        return label
    def fitValues(self): # returns (x,xexpr,y,yexpr,yerr,yerrexpr) or 6 None-s
        return (None, None, None, None, None, None)
    def setupFrame(self,noAxes=False):
        for widget in self.frame.winfo_children():
            widget.destroy()
        self.figure = Figure(tight_layout=True)
        self.canvas = FigureCanvasTkAgg(self.figure, self.frame)
        if self.topMessage is not None:
            tk.Label(self.frame,text=self.topMessage).pack(side=tk.TOP)
        self.canvas.get_tk_widget(). \
                        pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
        if not noAxes:
            self.axes = self.figure.add_subplot(111)
            self.canvas.mpl_connect('motion_notify_event', self.mouse_move)
            self.canvas.mpl_connect('button_press_event', self.mouse_click)
            self.canvas.mpl_connect('button_release_event', self.mouse_click)
            self.rs = RectangleSelector(self.axes, self.select, 
                        useblit=True, button=[1,], minspanx=5, minspany=5,
                        spancoords='pixels',state_modifier_keys={'clear':'shift'})
    # The RectangularSelector, select(), and mouse_click() implement a zoom
    # stack: click and drag with button 1 to zoom in, click button 3 to pop
    # the stack and zoom back out.
    # Also, mouse_move() displays the current mouse position in xpos and ypos.
    def select(self, edown, eup):
        # callback from RectangleSelector, called with 2 events.
        self.axesparam.xmin = min(edown.xdata,eup.xdata)
        self.axesparam.xmax = max(edown.xdata,eup.xdata)
        self.axesparam.ymin = min(edown.ydata,eup.ydata)
        self.axesparam.ymax = max(edown.ydata,eup.ydata)
        self.axesparam.autoscale.set(0)
        prevx = self.axes.get_xlim()
        prevy = self.axes.get_ylim()
        self.zoomStack.append( (prevx[0],prevy[0],prevx[1],prevy[1]) )
        self.draw()
    def mouse_click(self,event):
        if event.key is not None and  event.key == '\0':
            event.key = "shift"  # matplotlib BUG
        if event.key is not None and event.key.startswith("ctrl"):
            event.key = "control"
        # no zoom rectangle if any key is pressed (incl shift,control,alt,super)
        self.rs.set_active(event.key is None)
        if event.name == 'button_release_event':
            if event.button == 3 or (event.button == 1 and \
                                                    event.key == "control"):
                try:     # pop back to previous zoom
                    prev = self.zoomStack.pop()
                    self.axesparam.xmin = prev[0]
                    self.axesparam.ymin = prev[1]
                    self.axesparam.xmax = prev[2]
                    self.axesparam.ymax = prev[3]
                    self.axesparam.autoscale.set(0)
                    self.draw()
                except:
                    pass
        for i in self.mouse_click_callbacks:
            i(event)
    def mouse_move(self,event):
        x = event.xdata
        y = event.ydata
        if x == None:
            self.window.xpos.configure(text="")
        else:
            self.window.xpos.configure(text="    x: %.6g" % x)
        if y == None:
            self.window.ypos.configure(text="")
        else:
            self.window.ypos.configure(text="y: %.6g" % y)
        for i in self.mouse_move_callbacks:
            i(event)
    def toAPIString(self):
        return "#This unknown Plot is INVALID\n"
    def editStyle(self):
        class Edit(simpledialog.Dialog):
            def __init__(self,window,plot):
                self.window = window
                self.plot = plot
                simpledialog.Dialog.__init__(self,window,title="Edit Plot Style")
            def body(self,master):
                self.plot.style.insertEditor(master)
                self.plot.axesparam.insertEditor(master)
            def apply(self):
                self.plot.style.retrieve()
                self.plot.axesparam.retrieve()
                self.plot.draw()
        edit = Edit(self.window,self)

###
### SelectionPlot -- plot one column vs Time/N, and permit the user
### to select a range of data that applies to all subsequent plots.
### This is special, as it is integrated into the Input class; when a new
### window is created it gets a SelectionPlot automatically, with the
### columnName selected by the user. So there is no SelectionPlotDialog.
###
class SelectionPlot(Plot):
    def __init__(self,window, columnName=None, style=None, forceAxes=None):
        Plot.__init__(self,window,"Selection Plot",forceAxes)
        if columnName is not None:
            window.input.setSelectionPlotColumnName(columnName)
        self.columnName = window.input.getSelectionPlotColumnName()
        self.style = style if style is not None else Style()
        self.axesparam = AxesParam()
        self.selectMin = None
        self.selectMax = None
        self.selectStart = None
        self.prevLines = [None,None]
        self.type = "Selection"
        self.topMessage = "Click middle button and drag to select data, or shift-click and drag.\nData selection applies to all other plots in this window.\nClick left button and drag to zoom in; click right button to pop back out (control-click)."
        self.window.input.clearSelection()
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        axes.clear()
        self.selectStart = None
        self.prevLines = [None,None]
        data = self.window.input.getAllData()
        if data is None:
            data = {'N':[0.0,1.0]}
            print("NOTE: no data")
        if "Time" in data:
            x = data["Time"]
            xlabel = "Time"
        else:
            x = data["N"] # always present
            xlabel = "N"
        if self.columnName not in data:
            self.columnName = "N"
        y = data[self.columnName]
        self.name = "Selection Plot: " + self.columnName + " vs. " + xlabel
        self.axesparam.title = "Selection Plot"
        self.axesparam.xlabel = self.applyUnit(xlabel)
        self.axesparam.ylabel = self.applyUnit(self.columnName)
        x,y = self.thin(x,y)
        axes.plot(x, y, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,x,y)
        self.xArray = x
        self.xExpr = xlabel
        self.yArray = y
        self.yExpr = self.columnName
        # connect our own mouse handlers, for doing the selection
        self.register_move_callback(self.my_mouse_move)
        self.register_click_callback(self.my_mouse_click)
        # draw select lines
        if self.selectMin is not None and self.selectMax is not None:
            ybound = self.axes.get_ybound()
            self.prevLines[0] = self.axes.plot([self.selectMin,self.selectMin],
                                [ybound[0],ybound[1]],color="red")[0]
            self.prevLines[1]=self.axes.plot([self.selectMax,self.selectMax],
                                    [ybound[0],ybound[1]],color="red")[0]
        self.drawMarkers()    
        self.drawOutliers()
        self.callDrawCallbacks()
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def fitValues(self):
        data = self.window.input.getData()
        if data is None:
            data = {'N':[0.0,1.0]}
        if "Time" in data:
            x = data["Time"]
            xlabel = "Time"
        else:
            x = data["N"] # always present
            xlabel = "N"
        if self.columnName not in data:
            self.columnName = "N"
        y = data[self.columnName]
        return (x, xlabel, y, self.columnName, None, None)
    def my_mouse_click(self,event):
        if event.button == 2 or (event.button == 1 and event.key == "shift"):
            ybound = self.axes.get_ybound()
            self.axes.autoscale(enable=False)
            if event.name == 'button_press_event':
                for i in self.prevLines:
                    if i in self.axes.lines:
                        i.remove()
                self.prevLines = [None,None]
                self.prevLines[0] = self.axes.plot([event.xdata,event.xdata],
                                [ybound[0],ybound[1]],color="red")[0]
                self.selectStart = event.xdata
                self.selectStartPixels = event.x
            if event.name == 'button_release_event':
                if self.selectStart is not None:
                    if self.prevLines[1] in self.axes.lines:
                        self.prevLines[1].remove()
                    self.prevLines[1]=self.axes.plot([event.xdata,event.xdata],
                                    [ybound[0],ybound[1]],color="red")[0]
                    x0 = self.selectStart
                    x1 = event.xdata
                    if x0 > x1:
                        x1 = self.selectStart
                        x0 = event.xdata
                    self.selectStart = None
                    if abs(self.selectStartPixels-event.x) < 8:
                        self.window.input.clearSelection()
                        self.selectStartPixels = None
                        for i in self.prevLines:
                            if i in self.axes.lines:
                                i.remove()
                        self.selectMin = None
                        self.selectMax = None
                    else:
                        self.setSelectionRange(x0,x1)
                    self.window.redrawAll()
            self.canvas.draw()
    def my_mouse_move(self,event):
        if event.button == 2 or (event.button == 1 and event.key == "shift"):
            ybound = self.axes.get_ybound()
            if self.prevLines[1] in self.axes.lines:
                self.prevLines[1].remove()
            self.prevLines[1] = self.axes.plot([event.xdata,event.xdata],
                            [ybound[0],ybound[1]],color="red")[0]
            self.canvas.draw()
    def setSelectionRange(self,x0,x1):
        self.selectMin = x0
        self.selectMax = x1
        if "Time" in self.name:
            s = "(%g < Time) & (Time < %g)" % (x0,x1)
        else:
            s = "(%g < N) & (N < %g)" % (x0,x1)
        self.window.input.setSelection(s)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += "p%d = ap.SelectionPlot(win,'%s',style)\n" % \
                                        (self.number,self.columnName)
        if self.selectMin is not None and self.selectMax is not None:
            ret += "p%d.setSelectionRange(%g,%g)\n" % \
                                (self.number,self.selectMin,self.selectMax)
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret

###
### class SimpleDataPlot -- simple plot of one column vs time/N
###
class SimpleDataPlot(Plot):
    def __init__(self, window, columnName, style=None, forceAxes=None):
        Plot.__init__(self,window,"Data "+columnName,forceAxes)
        self.columnName = columnName
        self.type = "Simple Data"
        self.name = "Data "+columnName
        self.style = style if style is not None else Style()
        self.axesparam = AxesParam()
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        axes.clear()
        columnNames = self.window.input.columnNames()
        data = self.window.input.getData()
        if "Time" in columnNames:
            x = data["Time"]
            xlabel = "Time"
        else:
            x = data["N"]
            xlabel = "N"
        self.axesparam.xlabel = self.applyUnit(xlabel)
        self.axesparam.ylabel = self.applyUnit(self.columnName)
        y = data[self.columnName]
        x,y = self.thin(x,y)
        axes.plot(x, y, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,x,y)
        self.xArray = x
        self.xExpr = xlabel
        self.yArray = y
        self.yExpr = self.columnName
        self.drawMarkers()
        self.drawOutliers()
        self.callDrawCallbacks()
    def fitValues(self):
        return (self.xArray, self.xExpr, self.yArray, self.yExpr, None, None)
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += "p%d = ap.SimpleDataPlot(win,'%s',style)\n" % \
                                            (self.number,self.columnName)
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
class SimpleDataPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        simpledialog.Dialog.__init__(self,window,title="Simple Data Plot")
    def body(self,master):
        tk.Label(master,text="Simple Data Plot").pack()
        tk.Label(master,text="This plots one column of the data vs. either\nTime (if present) or N (index in the data).").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        ff = tk.Frame(master)
        tk.Label(ff,text="Column:").pack(side=tk.LEFT)
        self.col =  ttk.Combobox(ff,width=15,values=columnNames,
                                                state="readonly")
        self.col.pack(side=tk.LEFT)
        ff.pack()
        self.style.insertEditor(master)
    def apply(self):
        self.columnName = self.col.get()
        self.style.retrieve()
        SimpleDataPlot(self.window,self.columnName,self.style)

###
### class Histogram
###
class Histogram(Plot):
    def __init__(self,window,x,w="",cut="",nbins=100,errorbars=False,
                                    style=None,axesparam=None,forceAxes=None):
        Plot.__init__(self,window,"Histogram",forceAxes)
        self.x = x
        self.w = w
        self.cut = cut
        self.nbins = nbins
        self.errorbars = errorbars
        self.style = style if style is not None else Style()
        self.axesparam = axesparam if axesparam is not None else AxesParam()
        self.type = "Histogram"
        self.first = True
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        # evaluate the expressions
        data = self.window.input.getData()
        x = eval(self.x,Functions,data)
        if self.axesparam.xlabel == "": self.axesparam.xlabel =  \
                                                self.applyUnit(self.x)
        self.name = "Histogram "+self.x
        if self.w != "":
            w = eval(self.w,Functions,data)
            if getattr(w,"__len__",None) is None:
                v = w
                w = np.empty(len(x))
                w.fill(v)
            self.name += "  w="+self.w
        else:
            w = None
        if self.cut != "":
            cut = eval(self.cut,Functions,data)
            x = np.extract(cut,x)
            if w is not None:
                w = np.extract(cut,w)
            self.name += "  cut="+self.cut
        # no stat penel
        y,edges = np.histogram(x, bins=self.nbins, weights=w)
        yerr = None
        if self.errorbars:
            if w is not None:
                ww = w * w
                yerr,ignore = np.histogram(x, bins=self.nbins, weights=ww)
                yerr = np.sqrt(yerr)
            else:
                yerr = np.sqrt(y)
                # replace zero errorbars witn 1
                yerr[yerr < 1.0] = 1.0
            if self.first:
                self.style.linestyle = ' '
                self.style.pointMarker = '.'
                self.first = False
        centers = (edges[1:] + edges[:-1]) / 2.0
        self.xArray = centers
        self.yArray = y
        self.yerrArray = yerr
        drawstyle = "steps-mid"
        if self.style.pointMarker != ' ': drawstyle = None
        axes.errorbar(centers,y, yerr=yerr,
                drawstyle=drawstyle, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,edges,y)
        mean = np.mean(x)
        sigma = np.std(x)
        text = r'$N=%d, \mu=%.6g, \sigma=%.6g$' % (len(x), mean, sigma)
        axes.set_title(text,loc="right")
        axes.set_title("",loc="center")
        if self.axesparam.title == "":
            self.axesparam.title = self.cut
        axes.set_title(self.axesparam.title,loc="left")
        self.callDrawCallbacks()
    def fitValues(self):
        return (self.xArray, self.x, self.yArray, "(Histogram)",
                                        self.yerrArray, "sqrt(sum of w^2)")
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += self.axesparam.toAPIString()
        ret += "p%d = ap.Histogram(win,'%s','%s','%s',%d,%d,style,axesparam)\n" % \
            (self.number,self.x,self.w,self.cut,self.nbins,int(self.errorbars))
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
class HistogramDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        self.x = ""
        self.w = ""
        self.cut = ""
        self.nbins = 100
        self.errVar = tk.IntVar()
        self.errVar.set(0)
        self.errorbars = False
        simpledialog.Dialog.__init__(self,window,title="Histogram")
    def body(self,master):
        tk.Label(master,text="Histogram").pack()
        tk.Label(master,text="This plots a histogram of an expression; the expression\ncan involve fields (columns) of the data, and standard functions.").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        tk.Label(master,text="Enter expressions using the column names:"). \
                                                                pack()
        tk.Label(master,text=",".join(columnNames)).pack()
        paramNames = self.window.input.paramNames()
        if paramNames is not None and len(paramNames) > 0:
            tk.Label(master,text="And the parameter names:").pack()
            tk.Label(master,text=",".join(paramNames)).pack()
        tk.Label(master,text="In expressions: use & and | for and/or, parenthesize comparisons used with them.").pack()
        f = tk.Frame(master)
        tk.Label(f,text="X").grid(row=1,column=0)
        self.xEntry = tk.Entry(f,width=50)
        self.xEntry.grid(row=1,column=1)
        tk.Label(f,text="Weight").grid(row=2,column=0)
        self.wEntry = tk.Entry(f,width=50)
        self.wEntry.grid(row=2,column=1)
        tk.Label(f,text="Cut").grid(row=3,column=0)
        self.cutEntry = tk.Entry(f,width=50)
        self.cutEntry.grid(row=3,column=1)
        ff = tk.Frame(f)
        self.errCheckbutton = tk.Checkbutton(ff,
                    text="Statistical errorbars", variable=self.errVar)
        self.errCheckbutton.pack(side=tk.LEFT)
        tk.Label(ff,text="                N bins:").pack(side=tk.LEFT)
        self.nbinsEntry = tk.Entry(ff,width=8)
        self.nbinsEntry.pack(side=tk.LEFT)
        self.nbinsEntry.insert(tk.END,"100")
        ff.grid(row=5,column=1)
        f.pack()
        self.style.insertEditor(master)
        self.axesparam.insertEditor(master)
        return self.xEntry
    def validate(self):
        err = ""
        data = self.window.input.getData()
        x = self.xEntry.get().strip()
        if x == "":
            err += "Missing expression for X\n"
        if not validateExpr(x,data):
            err += "Invalid expression for X\n"
        w = self.wEntry.get().strip()
        if w != "" and not validateExpr(w,data):
            err += "Invalid expression for W\n"
        cut = self.cutEntry.get().strip()
        if cut != "" and not validateExpr(cut,data):
            err += "Invalid expression for Cut\n"
        if err != "":
            errorBox(err)
            return False
        return True
    def apply(self):
        x = self.xEntry.get().strip()
        w = self.wEntry.get().strip()
        cut = self.cutEntry.get().strip()
        errorbars = (self.errVar.get() != 0)
        try: nbins = int(self.nbinsEntry.get().strip())
        except: nbins = 100
        self.style.retrieve()
        self.axesparam.retrieve()
        Histogram(self.window,x,w,cut,nbins,errorbars,self.style,self.axesparam)

###
### class XYPlot
###
class XYPlot(Plot):
    def __init__(self, window, x, y, cut='', xerrbar='', yerrbar='', style=None,
                                                axesparam=None,forceAxes=None):
        Plot.__init__(self,window,"XYPlot",forceAxes)
        self.window = window
        self.x = x
        self.y = y
        self.cut = cut
        self.xerrbar = xerrbar
        self.yerrbar = yerrbar
        self.style = style if style is not None else Style()
        self.axesparam = axesparam if axesparam is not None else AxesParam()
        self.type = "XY Plot"
        self.first = True
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        if self.axesparam.xlabel == "": self.axesparam.xlabel = \
                                                    self.applyUnit(self.x)
        if self.axesparam.ylabel == "": self.axesparam.ylabel = \
                                                    self.applyUnit(self.y)
        # evaluate the expressions
        data = self.window.input.getData()
        x = eval(self.x,Functions,data)
        y = eval(self.y,Functions,data)
        self.name = "XY Plot y="+self.y+"  x="+self.x
        if self.cut != "":
            cut = eval(self.cut,Functions,data)
            x = np.extract(cut,x)
            y = np.extract(cut,y)
            self.axesparam.title = self.cut
            self.name += "  cut="+self.cut
        xerrbar = None
        if self.xerrbar != "":
            xerrbar = eval(self.xerrbar,Functions,data)
        yerrbar = None
        if self.yerrbar != "":
            yerrbar = eval(self.yerrbar,Functions,data)
        self.xArray = x
        self.yArray = y
        self.yerrArray = yerrbar
        if not self.isSorted(x) or xerrbar is not None or yerrbar is not None:
            if self.first:
                self.style.linestyle = ' '
                self.style.pointMarker = '.'
                self.first = False
        if self.axesparam.title == "":
            self.axesparam.title = self.cut
        x,y = self.thin(x,y)
        axes.errorbar(x, y, color=self.style.color, xerr=xerrbar, yerr=yerrbar,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,x,y)
        self.drawMarkers()
        self.drawOutliers()
        self.callDrawCallbacks()
    def fitValues(self):
        return (self.xArray, self.x, self.yArray, self.y, 
                                self.yerrArray, self.yerrbar)
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += self.axesparam.toAPIString()
        ret += "p%d = ap.XYPlot(win,'%s','%s','%s','%s','%s',style,axesparam)\n" % \
                (self.number,self.x,self.y,self.cut,self.xerrbar,self.yerrbar)
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
class XYPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        self.x = ""
        self.y = ""
        self.cut = ""
        simpledialog.Dialog.__init__(self,window,title="XYPlot")
    def body(self,master):
        tk.Label(master,text="XY Plot").pack()
        tk.Label(master,text="This plots two expressions vs. one another; the expressions\ncan involve the fields (columns) of the data, and standard functions.").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        tk.Label(master,text="Enter expressions using the column names:"). \
                                                                pack()
        tk.Label(master,text=",".join(columnNames)).pack()
        paramNames = self.window.input.paramNames()
        if paramNames is not None and len(paramNames) > 0:
            tk.Label(master,text="And the parameter names:").pack()
            tk.Label(master,text=",".join(paramNames)).pack()
        tk.Label(master,text="In expressions: use & and | for and/or, parenthesize comparisons used with them.").pack()
        f = tk.Frame(master)
        tk.Label(f,text="X").grid(row=1,column=0)
        self.xEntry = tk.Entry(f,width=50)
        self.xEntry.grid(row=1,column=1)
        tk.Label(f,text="Y").grid(row=2,column=0)
        self.yEntry = tk.Entry(f,width=50)
        self.yEntry.grid(row=2,column=1)
        tk.Label(f,text="Cut").grid(row=3,column=0)
        self.cutEntry = tk.Entry(f,width=50)
        self.cutEntry.grid(row=3,column=1)
        f.pack()
        f = tk.Frame(master)
        tk.Label(f,text="X errorbar:").pack(side=tk.LEFT)
        self.xerrEntry = tk.Entry(f,width=20)
        self.xerrEntry.pack(side=tk.LEFT)
        tk.Label(f,text="Y errorbar:").pack(side=tk.LEFT)
        self.yerrEntry = tk.Entry(f,width=20)
        self.yerrEntry.pack(side=tk.LEFT)
        f.pack()
        self.style.insertEditor(master)
        self.axesparam.insertEditor(master)
        return self.xEntry
    def validate(self):
        err = ""
        data = self.window.input.getData()
        x = self.xEntry.get().strip()
        if x == "":
            err += "Missing expression for X\n"
        if not validateExpr(x,data):
            err += "Invalid expression for X\n"
        y = self.yEntry.get().strip()
        if y == "":
            err += "Missing expression for Y\n"
        if not validateExpr(y,data):
            err += "Invalid expression for Y\n"
        cut = self.cutEntry.get().strip()
        if cut != "" and not validateExpr(cut,data):
            err += "Invalid expression for Cut\n"
        xerr = self.xerrEntry.get().strip()
        if xerr != "" and not validateExpr(xerr,data):
            err += "Invalid expression for X Errorbar\n"
        yerr = self.yerrEntry.get().strip()
        if yerr != "" and not validateExpr(yerr,data):
            err += "Invalid expression for Y Errorbar\n"
        if err != "":
            errorBox(err)
            return False
        return True
    def apply(self):
        x = self.xEntry.get().strip()
        y = self.yEntry.get().strip()
        cut = self.cutEntry.get().strip()
        xerr = self.xerrEntry.get().strip()
        yerr = self.yerrEntry.get().strip()
        self.style.retrieve()
        self.axesparam.retrieve()
        XYPlot(self.window,x,y,cut,xerr,yerr,self.style,self.axesparam)

###
### class FFTPlot
###
class FFTPlot(Plot):
    def __init__(self,window,x, sampleRate=1000, windowFunction=None, 
                                    style=None, axesparam=None, forceAxes=None):
        Plot.__init__(self,window,"FFT",forceAxes)
        self.x = x
        self.sampleRate = sampleRate
        self.windowFunction = windowFunction
        self.style = style if style is not None else Style()
        if axesparam is not None:
            self.axesparam = axesparam 
        else:
            self.axesparam = AxesParam()
            self.axesparam.logy.set(1)
        self.type = "FFT Plot"
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        # evaluate the expressions
        data = self.window.input.getData()
        x = eval(self.x,Functions,data)
        if "Time" in data and \
                        not self.checkSampleRate(data["Time"],self.sampleRate):
            messagebox.showwarning(title="Warning",
                message="Time samples are not uniform or sample rate is wrong")
        # compute the FFT
        self.name = "FFT x="+self.x
        self.n_fft = len(x)
        if self.n_fft > 1024*1024:
                self.n_fft = self.n_fft & ~4095 # significant speed-up
        self.n_fft = self.n_fft & ~31
        x = x[0:self.n_fft]
        print("calculating FFT:", self.n_fft, "data points")
        if self.windowFunction is not None:
            if isinstance(self.windowFunction,str): 
                self.windowFunction = self.windowFunction.lower()
            if self.windowFunction == 1 or self.windowFunction == "hamming":
                x *= np.hamming(self.n_fft)
            if self.windowFunction == 2 or self.windowFunction == "hanning":
                x *= np.hanning(self.n_fft)
        dft = fft.rfft(x, n=self.n_fft)
        norm = 2.0 / self.n_fft       # 2.0 because we omit negative freqs
        binWidth = float(self.sampleRate) / float(self.n_fft)
        print("binWidth=",binWidth," nfft=",self.n_fft," sampleRate=",self.sampleRate)
        fft_mod = np.abs(dft) * norm
        fft_freq = np.arange(len(dft)) * binWidth
        fft_freq,fft_mod = self.thin(fft_freq,fft_mod)
        axes.plot(fft_freq, fft_mod, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,fft_freq,fft_mod)
        self.callDrawCallbacks()
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += self.axesparam.toAPIString()
        ret += "p%d = ap.FFTPlot(win,'%s',%g,'%s',style,axesparam)\n" % \
                (self.number,self.x,self.sampleRate,self.windowFunction)
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
class FFTPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        self.axesparam.logy.set(1)
        self.axesparam.xlabel = "Frequency (Hz)"
        self.windowing = tk.IntVar()
        self.windowing.set(0)
        # compute sample rate, if possible
        data = self.window.input.getData()
        self.sampleRate = 1000
        if "Time" in data:
            dt = data["Time"][-1] - data["Time"][0]
            if dt > 0.0:
                self.sampleRate = round((len(data["Time"])-1) / dt, 1)
        simpledialog.Dialog.__init__(self,window,title="FFT")
    def body(self,master):
        tk.Label(master,text="FFT Plot").pack()
        tk.Label(master,text="This plots the Fast Fourier Transform of the data,\nwhich must be a regular time series.").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        tk.Label(master,text="Enter expression using the column names:"). \
                                                                pack()
        tk.Label(master,text=",".join(columnNames)).pack()
        f = tk.Frame(master)
        tk.Label(f,text="X").pack(side=tk.LEFT)
        self.xEntry = tk.Entry(f,width=50)
        self.xEntry.pack(side=tk.LEFT)
        f.pack()
        f = tk.Frame(master)
        tk.Label(f,text="Sample Rate").pack(side=tk.LEFT)
        self.samplerateEntry = tk.Entry(f,width=9)
        self.samplerateEntry.insert(0,str(self.sampleRate))
        self.samplerateEntry.pack(side=tk.LEFT)
        f.pack()
        f = tk.Frame(master)
        tk.Label(f,text="Window:").pack(side=tk.LEFT)
        tk.Radiobutton(f,text="None",variable=self.windowing,value=0). \
                                    pack(side=tk.LEFT)
        tk.Radiobutton(f,text="Hamming",variable=self.windowing,value=1). \
                                    pack(side=tk.LEFT)
        tk.Radiobutton(f,text="Hanning",variable=self.windowing,value=2). \
                                    pack(side=tk.LEFT)
        f.pack()
        self.style.insertEditor(master)
        self.axesparam.insertEditor(master)
        return self.xEntry
    def validate(self):
        err = ""
        data = self.window.input.getData()
        x = self.xEntry.get().strip()
        if x == "":
            err += "Missing expression for X\n"
        if not validateExpr(x,data):
            err += "Invalid expression for X\n"
        if err != "":
            errorBox(err)
            return False
        return True
    def apply(self):
        x = self.xEntry.get().strip()
        sampleRate = float(self.samplerateEntry.get().strip())
        self.style.retrieve()
        self.axesparam.retrieve()
        windowFunction = self.windowing.get()
        FFTPlot(self.window,x,sampleRate,windowFunction,self.style,
                                                                self.axesparam)

###
### class CorrelationPlot
###
class CorrelationPlot(Plot):
    def __init__(self,window,x1,x2,subtractMeans=False,extendX1=False,
                                    style=None,axesparam=None,forceAxes=None):
        Plot.__init__(self,window,"Correlation",forceAxes)
        self.x1 = x1
        self.x2 = x2
        self.autocorrelation = (x2 == "")
        if self.autocorrelation:
            self.x2 = self.x1
        self.subtractMeans = subtractMeans
        self.extendX1 = extendX1
        self.type = "Correlation Plot"
        self.style = style if style is not None else Style()
        self.axesparam = axesparam if axesparam is not None else AxesParam()
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        # evaluate the expressions and normaliza the data
        data = self.window.input.getData()
        x1data = eval(self.x1,Functions,data)
        x1data = x1data / np.sqrt(np.sum(x1data*x1data))
        if self.autocorrelation:
            x2data = x1data
        else:
            x2data = eval(self.x2,Functions,data)
            x2data = x2data / np.sqrt(np.sum(x2data*x2data))
        # optionally subtract means
        if self.subtractMeans:
            x1data = x1data - np.mean(x1data)
            x2data = x2data - np.mean(x2data)
        # optionally extend x1
        if self.extendX1:
            x1data = x1data.copy()
            x1data = np.append(x1data,x1data)
            mode = "valid"
        else:
            mode = "full"
        # ensure even # samples
        if len(x1data)&1: x1data = x1data[0:-1]
        if len(x2data)&1: x2data = x2data[0:-1]
        # compute the Correlation
        self.name = "Correlation x1="+self.x1+" x2="+self.x2
        print("calculating Correlation:", len(x1data), "data points")
        corr = np.correlate(x1data,x2data,mode=mode)
        # rotate if x1 was extended
        if self.extendX1:
            n = corr.size // 2
            corr = np.append(corr[n:],corr[1:n])
        # plot it
        self.axesparam.xlabel = "Samples offset"
        if self.autocorrelation:
            self.axesparam.ylabel = "Autocorrelation of " + self.x1
        else:
            self.axesparam.ylabel = "Correlation of " + self.x1 + \
                                    " with " + self.x2
        x = np.arange(corr.size) - corr.size//2
        axes.plot(x, corr, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.apply(axes,x,corr)
        self.callDrawCallbacks()
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += self.axesparam.toAPIString()
        ret += "p%d = ap.CorrelationPlot(win,'%s','%s',%d,%d,style,axesparam)\n" % \
                (self.number,self.x1,self.x2,int(self.subtractMeans),
                 int(self.extendX1))
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
class CorrelationPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        self.extendX1 = tk.IntVar()
        self.subtractMeans = tk.IntVar()
        simpledialog.Dialog.__init__(self,window,title="Correlation")
    def body(self,master):
        tk.Label(master,text="Correlation Plot").pack()
        tk.Label(master,text="""
This plots the correlation between two time series.
Each data series is normalized.
There will be end effects throughout, unless X1 is extended.
Leave X2 blank for autocorrelation plot.""").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        tk.Label(master,text="Enter expressions using the column names:"). \
                                                                pack()
        tk.Label(master,text=",".join(columnNames)).pack()
        f = tk.Frame(master)
        tk.Label(f,text="X1").pack(side=tk.LEFT)
        self.x1Entry = tk.Entry(f,width=50)
        self.x1Entry.pack(side=tk.LEFT)
        f.pack()
        f = tk.Frame(master)
        tk.Label(f,text="X2").pack(side=tk.LEFT)
        self.x2Entry = tk.Entry(f,width=50)
        self.x2Entry.pack(side=tk.LEFT)
        f.pack()
        tk.Checkbutton(master,text="Subtract means",
                                        variable=self.subtractMeans).pack()
        tk.Checkbutton(master,text="Extend X1 as if cyclical",
                                        variable=self.extendX1).pack()
        self.style.insertEditor(master)
        self.axesparam.insertEditor(master)
        return self.x1Entry
    def validate(self):
        err = ""
        data = self.window.input.getData()
        x1 = self.x1Entry.get().strip()
        if x1 == "":
            err += "Missing expression for x1\n"
        if not validateExpr(x1,data):
            err += "Invalid expression for x1\n"
        x2 = self.x2Entry.get().strip()
        if x2 != "":
            if not validateExpr(x2,data):
                err += "Invalid expression for x2\n"
        if err != "":
            errorBox(err)
            return False
        return True
    def apply(self):
        x1 = self.x1Entry.get().strip()
        x2 = self.x2Entry.get().strip()
        subtractMeans = (self.subtractMeans.get() != 0)
        extendX1 = (self.extendX1.get() != 0)
        self.style.retrieve()
        self.axesparam.retrieve()
        CorrelationPlot(self.window,x1,x2,subtractMeans,extendX1,
                                    self.style,self.axesparam)


###
### class AllanDevPlot
###
if Allantools:
  class AllanDevPlot(Plot):
    def __init__(self, window, x, sampleRate=None, style=None, axesparam=None,
                                                                forceAxes=None):
        #breakpoint()
        Plot.__init__(self,window,"AllanDev",forceAxes)
        self.x = x
        self.sampleRate = sampleRate
        self.style = style if style is not None else Style()
        self.axesparam = axesparam if axesparam is not None else AxesParam()
        self.type = "Allan Deviation"
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
            axes = self.axes
        if self.axesparam.ylabel == "":
            self.axesparam.ylabel = "Allan Deviation"
            self.axesparam.xlabel = "Time (seconds)"
        if self.axesparam.ylabel == "Allan Deviation":
            self.axesparam.ylabel += " of " + self.applyUnit(self.x)
        self.name = "Allan Deviation x="+self.x
        # evaluate the expressions
        data = self.window.input.getData()
        x = eval(self.x,Functions,data)
        if self.sampleRate is None:
            t = data["Time"]
            self.sampleRate = float(len(t) - 1) / (t[-1] - t[0])
        if "Time" in data and \
                        not self.checkSampleRate(data["Time"],self.sampleRate):
            messagebox.showwarning(title="Warning",
                message="Time samples are not uniform or sample rate is wrong")
        # compute the Allan Deviation -- considerable speed up due to 
        # not computing at all possible times.
        taus = []
        maxTime = float(len(x)-1) / float(self.sampleRate)
        t = 1.0/self.sampleRate
        f = 1.1
        while t < maxTime:
            taus.append(t)
            t *= f
        print("Computing Allan Deviation")
        print(x[0:10],x[-10:-1])
        (times, allanDev, ade, adn) = allantools.oadev(x, rate=self.sampleRate,
                                                taus=taus, data_type="freq")
        # special handling: Frequency => Picometers
        allanDev = self.handleFrequency2Picometers(allanDev)
        axes.plot(times, allanDev, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker,
                label=label)
        self.axesparam.logx.set(1)
        self.axesparam.logy.set(1)
        self.axesparam.apply(axes,times,allanDev)
        self.callDrawCallbacks()
    def handleFrequency2Picometers(self,allanDev):
        if self.x == "Frequency":
            scale = self.window.input.getParam("FrequencyScale",None)
            hzPerPm  = self.window.input.getParam("HzPerPicometer",None)
            if scale is not None and hzPerPm is not None:
                allanDev /= (hzPerPm * scale)
                self.axesparam.ylabel = "Allan Deviation (picometers)"
        return allanDev
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += self.style.toAPIString()
        ret += self.axesparam.toAPIString()
        ret += "p%d = ap.AllanDevPlot(win,'%s',%g,style,axesparam)\n" % \
                                    (self.number,self.x,self.sampleRate)
        if self.fit is not None:
            ret += self.fit.toAPIString()
        return ret
  class AllanDevPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        self.axesparam.logx.set(1)
        self.axesparam.logy.set(1)
        self.axesparam.xlabel = "Time (seconds)"
        self.axesparam.ylabel = "Allan Deviation"
        # compute sample rate, if possible
        data = self.window.input.getData()
        self.sampleRate = 1000
        self.timePresent = True
        if "Time" in data:
            dt = data["Time"][-1] - data["Time"][0]
            if dt > 0.0:
                self.sampleRate = round((len(data["Time"])-1) / dt, 1)
        else:
            self.timePresent = False
        simpledialog.Dialog.__init__(self,window,title="Alla Deviation")
    def body(self,master):
        if not self.timePresent:
            tk.Label(master,text="No Time present in the input; Allan Deviation is not possible.").pack()
            return
        tk.Label(master,text="Allan Deviation Plot").pack()
        tk.Label(master,text="This plots the Allan Deviation of the data,\n which must be a regular time series.").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        tk.Label(master,text="Enter expression using the column names:"). \
                                                                pack()
        tk.Label(master,text=",".join(columnNames)).pack()
        f = tk.Frame(master)
        tk.Label(f,text="X").pack(side=tk.LEFT)
        self.xEntry = tk.Entry(f,width=50)
        self.xEntry.pack(side=tk.LEFT)
        f.pack()
        f = tk.Frame(master)
        tk.Label(f,text="Sample Rate").pack(side=tk.LEFT)
        self.samplerateEntry = tk.Entry(f,width=9)
        self.samplerateEntry.insert(0,str(self.sampleRate))
        self.samplerateEntry.pack(side=tk.LEFT)
        f.pack()
        self.style.insertEditor(master)
        self.axesparam.insertEditor(master)
        return self.xEntry
    def validate(self):
        err = ""
        data = self.window.input.getData()
        x = self.xEntry.get().strip()
        if x == "":
            err += "Missing expression for X\n"
        if not validateExpr(x,data):
            err += "Invalid expression for X\n"
        if err != "":
            errorBox(err)
            return False
        return True
    def apply(self):
        if not self.timePresent: return
        x = self.xEntry.get().strip()
        sampleRate = float(self.samplerateEntry.get().strip())
        self.style.retrieve()
        self.axesparam.retrieve()
        AllanDevPlot(self.window,x,sampleRate,self.style,self.axesparam)
else:
  class AllanDevPlot(Plot):
    def __init__(self, window, x, sampleRate=None, style=None, axesparam=None,
                                                                forceAxes=None):
        raise NotImplementedError("AllenDevPlot is not available, as the allantools are not installed")

###
### class QuadPlot
###
class QuadPlot(Plot):
    def __init__(self, window, columnName, forceAxes=None):
        Plot.__init__(self,window,"Quad",forceAxes)
        self.columnName = columnName
        self.type = "Quad Plot"
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame(True)
        self.name = "Quad Plot"
        axes1 = self.figure.add_subplot(221)
        SimpleDataPlot(self.window,self.columnName,forceAxes=axes1)
        axes2 = self.figure.add_subplot(222)
        FFTPlot(self.window,self.columnName,forceAxes=axes2)
        axes3 = self.figure.add_subplot(223)
        Histogram(self.window,self.columnName,forceAxes=axes3)
        axes4 = self.figure.add_subplot(224)
        AllanDevPlot(self.window,self.columnName,forceAxes=axes4)
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += "p%d = ap.QuadPlot(win,'%s')\n" % (self.number,self.columnName)
        return ret
class QuadPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.columnName = ""
        simpledialog.Dialog.__init__(self,window,title="Quad Plot")
    def body(self,master):
        tk.Label(master,text="Quad Plot").pack()
        tk.Label(master,text="This plots four useful plots of a regular time series:\nData, FFT, Histogram, and Allan Deviation.").pack()
        tk.Label(master,text="").pack()
        columnNames = self.window.input.columnNames()
        ff = tk.Frame(master)
        tk.Label(ff,text="Column:").pack(side=tk.LEFT)
        self.col =  ttk.Combobox(ff,width=15,values=columnNames,
                                                state="readonly")
        i = 0
        if columnNames[i] == "Time": i = 1
        self.col.current(i)
        self.col.pack(side=tk.LEFT)
        ff.pack()
    def validate(self):
        return True
    def apply(self):
        columnName = self.col.get()
        QuadPlot(self.window,columnName)

###
### class ComparePlot
###
class ComparePlot(Plot):
    def __init__(self, window, plotList, commonX=False, xmin=None, xmax=None,
                                    forceAxes=None):
        Plot.__init__(self,window,"Compare",forceAxes)
        self.plotList = plotList
        self.commonX = commonX
        self.xmin = xmin if xmin is not None else 0.0
        self.xmax = xmax if xmin is not None else 1.0
        self.type = "Compare Plot"
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame(True)
        self.name = "Compare Plot"
        nrows = len(self.plotList)
        for i in range(nrows):
            ap = self.plotList[i].axesparam
            t1,t2 = (ap.xmin,ap.xmax)
            ax = self.figure.add_subplot(nrows,1,i+1)
            if self.commonX:
                if i < nrows-1:
                    # for all but the bottom plot, hide X axis ticks and label
                    ax.set_xticklabels([])
                    ax.get_xaxis().get_label().set_visible(False)
                # ap.autoscale is already cleared, as it applies only once
                ap.xmin = self.xmin
                ap.xmax = self.xmax
            self.plotList[i].draw(ax)
            ap.xmin,ap.xmax = (t1,t2)
            ax.set_title(self.plotList[i].name,loc='left')
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += "p%d = ap.ComparePlot(win,[" % self.number
        for p in self.plotList:
            ret += "p%d," % p.number
        ret += "],%d,%g,%g)\n" % (int(self.commonX),self.xmin,self.xmax)
        return ret
class ComparePlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.plots = []
        self.names = []
        self.commonVar = tk.IntVar()
        self.commonVar.set(0)
        self.xmin = None
        self.xmax = None
        simpledialog.Dialog.__init__(self,window,title="Compare Plot")
    def body(self,master):
        tk.Label(master,text="Compare Plot").pack()
        tk.Label(master,text="This draws one to eight existing plots on one screen for comparison.\nThey are stacked vertically, so presumably their X axes are related.").pack()
        tk.Label(master,text="").pack()
        self.allPlots = self.window.listAllPlots()
        self.allNames = [re.sub(".*/","",p[0])+":"+p[1] \
                                                for p in self.allPlots]
        self.sel = []
        for i in range(8):
            sel = ttk.Combobox(master,width=40,values=self.allNames,
                                        justify='left',state="readonly")
            sel.pack()
            self.sel.append(sel)
        tk.Label(master,text="Compare Plot Style").pack()
        tk.Label(master,text="Plots are drawn with their own styles.\n" + \
            "This editor can only control the common X axis scale.\n" + \
            "\nYou can change the style of any plot, then select this\n " +
            "plot and do Plot/Redraw to display the new style.").pack()
        tk.Checkbutton(master,text="Force Common X limits",
                                    variable=self.commonVar).pack()
        f = tk.Frame(master)
        tk.Label(f,text="X Min:").pack(side=tk.LEFT)
        self.xminEntry = tk.Entry(f,width=8)
        self.xminEntry.pack(side=tk.LEFT)
        if self.xmin is not None:self.xminEntry.insert(0,"%.8g" % self.xmin)
        tk.Label(f,text="X Max:").pack(side=tk.LEFT)
        self.xmaxEntry = tk.Entry(f,width=8)
        self.xmaxEntry.pack(side=tk.LEFT)
        if self.xmax is not None:self.xmaxEntry.insert(0,"%.8g" % self.xmax)
        f.pack()
    def validate(self):
        self.plots = []
        self.names = []
        for i in range(8):
            j = self.sel[i].current()
            if j < 0 or j >= len(self.allPlots): continue
            self.plots.append(self.allPlots[j][2])
            self.names.append(self.allNames[j])
        if self.commonVar.get() != 0:
            self.commonX = True
            s = self.xminEntry.get().strip()
            if s != "": self.xmin = float(s)
            s = self.xmaxEntry.get().strip()
            if s != "": self.xmax = float(s)
        else:
            self.commonX = False
        return len(self.plots) > 0
    def apply(self):
        commonX = (self.commonVar.get() != 0)
        xmin = self.xmin
        xmax = self.xmax
        ComparePlot(self.window,self.plots,commonX,xmin,xmax)

###
### class ComboPlot
###
class ComboPlot(Plot):
    def __init__(self,window,plotList,forceAxes=None):
        Plot.__init__(self,window,"Combo",forceAxes)
        self.plotList = plotList
        self.type = "Combo Plot"
        self.style = Style()
        self.axesparam = AxesParam()
        self.draw(forceAxes)
    def draw(self,axes=None, label=None):
        if axes is None:
            self.setupFrame()
        else:
            self.axes = axes
        self.name = "Combo Plot"
        x = None
        y = None
        for plot in self.plotList:
                label = plot.axes.get_title(loc='center')
                if label is None or label == '':
                    label = plot.axes.get_title(loc='left')
                plot.draw(self.axes,label)
                if x is None:
                    x = [plot.axesparam.xmin, plot.axesparam.xmax]
                    y = [plot.axesparam.ymin, plot.axesparam.ymax]
                else:
                    x = [min(x[0],plot.axesparam.xmin),
                                        max(x[1],plot.axesparam.xmax)]
                    y = [min(y[0],plot.axesparam.ymin),
                                        max(y[1],plot.axesparam.ymax)]
        self.axesparam.apply(self.axes,x,y)
        self.axes.set_title("",loc="right")
        self.axes.set_title("",loc="left")
        self.axesparam.apply(self.axes,x,y)
        self.drawMarkers()
        self.drawOutliers()
        self.axes.legend()
    def reload(self):
        self.axesparam.autoscale.set(1)
        self.draw()
    def toAPIString(self):
        ret = "\n"
        ret += "p%d = ap.ComboPlot(win,[" % self.number
        for p in self.plotList:
            ret += "p%d," % p.number
        ret += "])\n"
        return ret
class ComboPlotDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        self.style = Style()
        self.axesparam = AxesParam()
        simpledialog.Dialog.__init__(self,window,title="Combo Plot")
    def body(self,master):
        tk.Label(master,text="Combo Plot").pack()
        tk.Label(master,text="This draws multiple existing plots on a single (common) set of axes.\nIt uses each plot's title in the legend.").pack()
        tk.Label(master,text="").pack()
        tk.Label(master,text="Select Plots to display:").pack()
        self.allPlots = self.window.listAllPlots()
        self.allNames = [re.sub(".*/","",p[0])+":"+p[1] \
                                                for p in self.allPlots]
        self.incl = []
        for i in range(len(self.allNames)):
            j = tk.IntVar()
            j.set(0)
            tk.Checkbutton(master,text=self.allNames[i],variable=j). \
                                        pack(anchor="w")
            self.incl.append(j)
        self.axesparam.insertEditor(master)
    def validate(self):
        return True
    def apply(self):
        plotList = []
        for i in range(len(self.incl)):
            if self.incl[i].get() != 0:
                plotList.append(self.allPlots[i][2])
        self.axesparam.retrieve()
        ComboPlot(self.window,plotList)

