#   root.py - input routines for Root

import sys, time, re, os, math
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox
from tkinter import simpledialog
import numpy as np
from copy import *

# let it raise an exception if not installed
# note that the "pyroot" module can have two different names.
try:
    import ROOT as root
except:
    import pyroot as root

root.gROOT.SetBatch(True)

from input import Input

###
### getListOfTNtuplePaths() - get a list of pathnames for all TNtuple-s
### in a Root TFile.
###
def getListOfTNtuplePaths(tfile,path='/',ret=[]):
    for key in tfile.GetListOfKeys():
        name  = key.GetName()
        obj = tfile.Get(name)
        if obj.IsA().InheritsFrom('TDirectory'):
            getListOfTNtuplePaths(obj,path+name+'/',ret)
        elif obj.IsA().InheritsFrom('TNtuple'):
            ret.append(path+name)
    return ret

###
### getColumnNamesFromTNtuplePath() - get the column (field) names for
### the TNtuple..
###
def getColumnNamesFromTNtuplePath(tfile,path):
    return [j.GetName() for j in tfile.Get(path).GetListOfBranches()]

###
### getDataFromTNtuplePath() - reads a TNtuple from the Root TFile,
### returning a dictionary mapping each field name to a 1-D numpy array.
###
def getDataFromTNtuplePath(tfile,path):
    ntuple = tfile.Get(path)
    # Root < v6.26 uses AsMatrix(), but >= V6.26 uses RDataFrame.AsNumpy()
    if hasattr(ntuple,"AsMatrix"):
        matrix, labels = ntuple.AsMatrix(return_labels=True)
        ret = {}
        for i in range(len(labels)):
            ret[labels[i]] = matrix[:,i].astype(np.float32)
        return ret
    else:
        return root.RDataFrame(path,tfile).AsNumpy()

###
### class RootInputParameters
###
class RootInputParameters():
    def isValid(self):
        return True
    def addTab(self,tabs):
        self.parent = tabs.master
        self.tfile = None
        self.input = None
        self.filename = ""
        f = tk.Frame(tabs)
        tabs.add(f,text="Root File",padding="0.1i")
        tk.Label(f,text="Root File Input").pack()
        tk.Label(f,text="").pack()
        tk.Button(f,text="Select Root File",command=self.selectRootFile). \
                                                pack()
        self.rootFilename = tk.Label(f,text="")
        self.rootFilename.pack(fill=tk.X)
        tk.Label(f,text="Then select a TNtuple from it:").pack()
        self.rootTNtuple = ttk.Combobox(f,width=30,state='readonly',
                                                        values=[])
        self.rootTNtuple.bind("<<ComboboxSelected>>",self.getColumnNames)
        self.rootTNtuple.pack()
        tk.Label(f,text=" ").pack()
        tk.Label(f,text="A Selection Plot will be displayed showing" + \
                                " one column:").pack()
        ff = tk.Frame(f)
        self.col = ttk.Combobox(ff,width=15,values=[],state="readonly")
        self.col.pack(side=tk.LEFT)
        ff.pack()
        tk.Label(f,text="Then other plots can be displayed.").pack()
        tk.Label(f,text=" ").pack()
    def getInput(self):
        self.input = RootInputFile(self)
        self.input.setSelectionPlotColumnName(self.col.get())
        return self.input
    def selectRootFile(self):
        self.col.configure(values=[])
        self.col.set("")
        self.filename = filedialog.askopenfilename(parent=self.parent)
        if self.filename is None: self.filename = ""
        if not self.filename == "":
            self.tfile = root.TFile(self.filename)
            if self.tfile.IsZombie():
                messagebox.showerror(title="AllPlot Error",
                            message="Cannot open Root file '"+self.filename+"'")
                self.filename = ""
                self.tfile.close()
                self.tfile = None
            else:
                self.tntuplenames = getListOfTNtuplePaths(self.tfile,"/",[])
                if len(self.tntuplenames) == 0:
                    messagebox.showerror(title="AllPlot Error",
                            message="Invalid Root file (no TNtuples) '"+self.filename+"'")
                    self.filename = ""
                    self.tfile.close()
                    self.tfile = None
                else:
                    self.rootTNtuple.config(values=self.tntuplenames)
        self.rootFilename.configure(text=self.filename)
    def getColumnNames(self,eventIgnored):
        self.ntuplePath = self.rootTNtuple.get()
        try:
            self.colNames = getColumnNamesFromTNtuplePath(self.tfile,
                                        self.ntuplePath)
            colnames = copy(self.colNames)
            colnames.insert(0,"(none)")
            self.col.configure(values=colnames)
            self.col.current(1)
        except:
            pass

###
### class RootInputFile - read a Root TFile and select a TNtuple in it
###
class RootInputFile(Input):
    def __init__(self,paramOrFilename,ntuplePath=None):
        Input.__init__(self)
        if isinstance(paramOrFilename,str):
            self.fn = paramOrFilename
            self.param = None
            self.tfile = root.TFile(self.fn)
            if self.tfile.IsZombie():
                print("ERROR: Cannot open Root file '"+self.fn+"'")
                messagebox.showerror(title="AllPlot Error",
                        message="Cannot open Root file '"+self.fn+"'")
                self.tfile.close()
                self.tfile = None
                self.filename = ""
                self.shortFIlename = ""
                self.fn = ""
                self.ntuplePath = None
            else:
                self.filename = self.fn+":"+ntuplePath
                self.shortFilename = os.path.basename(self.filename)
                self.ntuplePath = ntuplePath
                self.colNames = getColumnNamesFromTNtuplePath(self.tfile
                                                                ,ntuplePath)
            self.data = None
        else:
            self.param = paramOrFilename
            self.tfile = self.param.tfile
            self.ntuplePath = self.param.ntuplePath
            self.filename = self.param.filename+":"+self.param.ntuplePath
            self.shortFilename = os.path.basename(self.filename)
            self.fn = self.param.filename
            self.colNames = self.param.colNames
            self.data = None
    def read(self):
        save = self.saveAllParams()
        self.data = getDataFromTNtuplePath(self.tfile,self.ntuplePath)
        self.data["N"] = np.arange(len(self.data[self.colNames[0]]))
        self.restoreAllParams(save)
    def toAPIString(self):
        ret = "\n"
        ret += "inp = ap.RootInputFile('%s','%s')\n" % (self.fn,self.ntuplePath)
        if self.selectionPlotColumnName is not None:
            ret += "inp.setSelectionPlotColumnName('%s')\n" % \
                                                self.selectionPlotColumnName
        return ret

