#   textinput.py

import sys, time, re, os, math
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox
from tkinter import simpledialog
import numpy as np
from copy import *
import pathlib

#   pandas.read_csv() is MUCH faster than numpy.loadtxt()
try:
    import pandas as pd
    UsePandas = True
except:
    UsePandas = False
    print("Note: installing pandas will greatly speed up reading text files:")
    print("    pip3 install pandas")

from input import InputParameters, Input

###
### class TextInputParameters
###
class TextInputParameters(InputParameters):
    def isValid(self):
        return True
    def addTab(self,tabs):
        self.parent = tabs.master
        self.input = None
        f = tk.Frame(tabs)
        tabs.add(f,text="Text File",padding="0.1i")
        tk.Label(f,text="Text File Input").pack()
        tk.Label(f,text="").pack()
        tk.Button(f,text="Select Text File",command=self.selectTextFile). \
                                                pack()
        self.textFilename = tk.Label(f,text="")
        self.textFilename.pack(fill=tk.X)
        ff = tk.Frame(f)
        self.generateTimeVar = tk.IntVar()
        self.generateTimeVar.set(0)
        self.generateTime = tk.Checkbutton(ff,
                                text="Generate Time, sample rate:",
                                variable=self.generateTimeVar)
        self.generateTime.pack(side=tk.LEFT)
        self.sampleRate = tk.Entry(ff,width=6)
        self.sampleRate.insert(tk.END,"1000")
        self.sampleRate.pack(side=tk.LEFT)
        tk.Label(ff,text=" Hz").pack(side=tk.LEFT)
        ff.pack()
        tk.Label(f,text=" ").pack()
        tk.Label(f,text="A Selection Plot will be displayed showing" + \
                                " one column:").pack()
        ff = tk.Frame(f)
        self.col = ttk.Combobox(ff,width=15,values=[],state="readonly")
        self.col.pack(side=tk.LEFT)
        ff.pack()
        tk.Label(f,text="Then other plots can be displayed.").pack()
        tk.Label(f,text=" ").pack()
        ff = tk.Frame(f)
        tk.Label(ff,text="Limit events:").pack(side=tk.LEFT)
        self.minEvent = tk.Entry(ff,width=10)
        self.minEvent.insert(tk.END,"0")
        self.minEvent.pack(side=tk.LEFT)
        tk.Label(ff,text="to:").pack(side=tk.LEFT)
        self.maxEvent = tk.Entry(ff,width=10)
        self.maxEvent.insert(tk.END,"10000000")
        self.maxEvent.pack(side=tk.LEFT)
        ff.pack()
        tk.Label(f,text=" ").pack()
        self.doublePrecision = tk.IntVar()
        self.doublePrecision.set(0)
        tk.Checkbutton(f,text="Use Double Precision (rarely needed)",
                                variable=self.doublePrecision).pack()
    def getInput(self):
        if self.input is None: return None
        if self.generateTimeVar.get() != 0:
            self.input.setSampleRate(int(self.sampleRate.get()))
        self.input.setEventLimits(int(self.minEvent.get()),
                                    int(self.maxEvent.get()))
        self.input.setSelectionPlotColumnName(self.col.get())
        if self.doublePrecision.get() != 0: self.input.dtype = np.float64
        return self.input
    def selectTextFile(self):
        self.generateTime.configure(state=tk.NORMAL)
        self.sampleRate.configure(state=tk.NORMAL)
        self.col.configure(values=[])
        self.col.set("")
        filename = filedialog.askopenfilename(parent=self.parent)
        if filename is not None:
            filename = os.path.abspath(filename)
            self.input = TextInputFile(filename)
            if self.input.isValid():
                colnames = copy(self.input.colNames)
                if colnames.count("Time") > 0:
                    colnames.remove("Time")
                    self.generateTimeVar.set(0)
                    self.generateTime.configure(state=tk.DISABLED)
                    self.sampleRate.configure(state=tk.DISABLED)
                colnames.insert(0,"(none)")
                self.col.configure(values=colnames)
                self.col.current(1)
            else:
                messagebox.showerror(title="AllPlot Error",
                        message="Invalid text input file '"+filename+"'")
                filename = ""
        else:
            filename = ""
        self.textFilename.configure(text=filename)

###
### class TextInputFile - read a text file
###
class TextInputFile(Input):
    def __init__(self, filename, minEvent=0, maxEvent=1000000):
        Input.__init__(self)
        filename = os.path.abspath(filename)
        print("TextInputFile(%s)" % filename)
        self.filename = filename
        self.shortFilename = os.path.basename(filename)
        self.minEvent = minEvent
        self.maxEvent = maxEvent
        self.firstDataPos = 0
        self.dtype = np.float32
        self.colNames = None
        self.nColumns = 0
        self.separator = None
        comments = []
        try:
            self.inp = open(self.filename,"r")
        except:
            self.inp = None
            return
        # read lines until first data line, look for column names
        while True:
            self.firstDataPos = self.inp.tell()
            line = self.inp.readline()
            if line is None or line == "":
                # EOF with no data lines - file is not valid
                self.inp.close()
                self.inp = None
                return
            line = line.strip()
            if len(line) == 0: continue # ignore blank lines
            if line.startswith("#") or line.startswith("%"):
                comments.append(line[1:].strip())
            else: # this is the first non-comment line
                if line.startswith('"') or line.startswith("'"):
                    comments.append(line) # CSV column names in quotes
                    continue
                self.getColumnNames(comments,line,r'\s+') # space or tab
                if self.colNames is None:
                    self.getColumnNames(comments,line,r',') # comma
                break
        if self.nColumns == 0:
            # could not count columns
            self.inp.close()
            self.inp = None
            return
        if self.colNames is None:   # found no column names, generate them
            self.colNames = []
            for i in range(self.nColumns):
                self.colNames.append("c"+str(i+1))
        self.inp.seek(self.firstDataPos)
    def getColumnNames(self,comments,firstDataLine,separator):
        cols = re.split(separator,firstDataLine)
        try:  # check each column is a float
            for c in cols:
                tmp = float(c)  # initial and final spaces are ignored
        except:
            return
        self.nColumns = len(cols)
        for com in comments:    # initial # or % already removed from each line
            if com.startswith("#"): continue   # ## is not column names
            cols = re.split(separator,com)
            if len(cols) == self.nColumns:
                ok = True
                # remove any quotes, check valid identifier
                for i in range(len(cols)):
                    # Excel sometimes triples the quotes
                    while cols[i].startswith('"') and cols[i].endswith('"'):
                        cols[i] = cols[i][1:-1]
                    while cols[i].startswith("'") and cols[i].endswith("'"):
                        cols[i] = cols[i][1:-1]
                    if not re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$',cols[i]):
                        ok = False
                        break
                if not ok: continue
                self.colNames = cols
                self.separator = separator
                return
    def isValid(self):
        return self.colNames is not None
    def read(self):
        if self.inp is None:
            try:
                self.inp = open(self.filename,"r")
                self.inp.seek(self.firstDataPos)
            except:
                print("ERROR cannot reopen "+self.filename)
                return
        save = self.saveAllParams()
        self.data = {}
        self.restoreAllParams(save)
        # save file modification time
        self.st_mtime = pathlib.Path(self.filename).stat().st_mtime
        # clean up in case this is a reload
        try:
            self.colNames.remove("N")
        except:
            pass
        try:
            if self.sampleRate > 0: self.colNames.remove("Time")
        except:
            pass
        # read the file
        if UsePandas:
            if self.separator is None:
                self.separator = '\s+'
            dataset = pd.read_csv(self.inp,delimiter=self.separator,
                        comment='#', header=None,skiprows=self.minEvent,
                        nrows=self.maxEvent-self.minEvent+1, dtype=self.dtype,
                        names=self.colNames,engine='c')
            for n in self.colNames:
                self.data[n] = dataset[n].to_numpy()
        else:
            # construct a command to read the file, and exec() it
            vars = {'d':self.data, 'inp':self.inp, 'minEvent':self.minEvent,
                    'maxEvent':self.maxEvent, 'loadtxt':np.loadtxt, 
                    'dtype':self.dtype}
            cmd = ""
            for i in self.colNames:
                cmd += "d['"+i+"'],"
            cmd = cmd[:-1] + " = loadtxt(inp,unpack=True,dtype=dtype,skiprows=minEvent, max_rows=maxEvent-minEvent+1)"
            exec(cmd,vars)
        # loadtxt/read_csv closed the self.inp file
        self.inp  = None
        # always add N; add Time if sampleRate>0, and apply frequencyFactor
        self.data["N"] = np.arange(len(self.data[self.colNames[0]]))
        self.colNames.append("N")
        if self.sampleRate != 0:
            self.generateTimeColumn(self.sampleRate)
        if self.frequencyFactor != 0 and  "Frequency" in self.data:
            self.data["Frequency"] *= self.frequencyFactor
    def isChanged(self):
        try:
            st_mtime = pathlib.Path(self.filename).stat().st_mtime
            if st_mtime > self.st_mtime:
                return True
        except: 
            pass
        return False
    def toAPIString(self):
        ret = "\n"
        ret += "inp = ap.TextInputFile('%s',%d,%d)\n" % \
                        (self.filename,self.minEvent,self.maxEvent)
        if self.sampleRate != 0:
            ret += "inp.generateTimeColumn(%g)\n" % self.sampleRate
        if self.frequencyFactor != 0 and  "Frequency" in self.data:
            ret += "inp.setFrequencyFactor(%g)\n" % self.frequencyFactor
        return ret

