#   tools.py - tools applied to an input or plot.

import sys, time, re, os, math
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import simpledialog
from tkinter import scrolledtext
import tkinter.filedialog as filedialog
import os.path as path
import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import numpy as np
from copy import copy

import plot

try:
    from scipy.optimize import minimize
    FitAvailable = True
except:
    FitAvailable = False

try:
    import iit
    IIT_present = True
    print("IIT extensions are present")
except:
    IIT_present = False

###
### setupToolMenu()
###
def setupToolMenu(window):
    global FitAvailable
    toolmenu = tk.Menu(window.menubar)
    window.menubar.add_cascade(label="Tools", menu=toolmenu)
    if FitAvailable:
        toolmenu.add_command(label="Fit Panel",
                command=lambda: displayFitPanel(window))
    toolmenu.add_command(label="Markers",
                command=lambda: displayMarkerPanel(window))
    toolmenu.add_command(label="Parameters",
                command=lambda: editParameters(window))
    toolmenu.add_separator()
    toolmenu.add_command(label="Create Column",
                command=lambda: createColumn(window))
    toolmenu.add_separator()
    toolmenu.add_command(label="Histogram Sample Intervals",
                command=lambda: histogramSampleIntervals(window))
    toolmenu.add_command(label="Generate DeltaTime",
                command=lambda: generateDeltaTime(window))
    if IIT_present:
        iit.setupToolMenu(window,toolmenu)

###
### class FitPanel -- control fitting to the current plot; it is  a non-modal
### dialog that places itself to the right of the current plot window.
###
class FitPanel(tk.Toplevel):
    def __init__(self,window,plot):
        tk.Toplevel.__init__(self,window)
        self.window = window
        self.plot = plot
        self.plot.fitPanel = self
        self.fit = None
        self.protocol("WM_DELETE_WINDOW", self.windowClosed)
        xpos = window.winfo_x() + window.winfo_width()
        ypos = window.winfo_y()
        self.geometry("+%d+%d" % (xpos,ypos))
        self.title("Fit Panel "+self.plot.name)
        self.bind('<Return>', lambda event: self.doFit())
        tk.Label(self,text="""
This fit panel will fit a parameterized function f(x) to
the current plot interpreted as y(x). The function can use any
existing parameters, and/or new ones defined below.""").pack()
        self.x,self.xExpr,self.y,self.yExpr,self.yerr,self.yerrExpr = \
                                                self.plot.fitValues()
        tk.Label(self,text="x: "+self.xExpr,anchor=tk.W). \
                                    pack(fill=tk.X)
        tk.Label(self,text="y: "+self.yExpr,anchor=tk.W). \
                                    pack(fill=tk.X)
        if self.yerr is not None:
            tk.Label(self,text="y errorbars: "+self.yerrExpr,anchor=tk.W). \
                                    pack(fill=tk.X)
        self.ndataL = tk.Label(self,text="",anchor=tk.W)
        self.ndataL.pack(fill=tk.X)
        paramNames = window.input.paramNames()
        if len(paramNames) > 0:
            tk.Label(self,text="parameters: "+",".join(paramNames),
                                    anchor=tk.W).pack(fill=tk.X)
        tk.Label(self,text="The function to be fit (minimizing Chi-sq):"). \
                                                pack(fill=tk.X)
        f = tk.Frame(self)
        tk.Label(f,text="f(x)=").pack(side=tk.LEFT)
        self.fExprE = tk.Entry(f)
        self.fExprE.pack(fill=tk.X)
        f.pack(fill=tk.X)
        f = tk.Frame(self)
        tk.Label(f,text="Minimization method:").pack(side=tk.LEFT)
        m = [ 'Nelder-Mead', 'Powell', 'CG', 'BFGS', 'L-BFGS-B',
               'TNC', 'COBYLA', 'SLSQP' ]
        self.methodE = ttk.Combobox(f, values=m, height=len(m), width=12,
                                                        state='readonly')
        self.methodE.current(0)
        self.methodE.bind("<<ComboboxSelected>>",self.methodChanged)
        self.methodE.pack(side=tk.LEFT)
        f.pack(fill=tk.X)
        f = tk.Frame(self)
        tk.Label(f,text="Parameters to be varied (new or existing):"). \
                                                        pack(side=tk.LEFT)
        tk.Button(f,text="Plot Current Values",command=self.plotFit). \
                                                        pack(side=tk.RIGHT)
        f.pack(fill=tk.X)
        f = tk.Frame(self)
        tk.Label(f,text="Name").grid(row=0,column=0)
        tk.Label(f,text="Value").grid(row=0,column=1)
        tk.Label(f,text="Fixed").grid(row=0,column=2)
        self.stepL = tk.Label(f,text="Initial Step")
        self.stepL.grid(row=0,column=3)
        self.minL = tk.Label(f,text="Minimum")
        self.minL.grid(row=0,column=4)
        self.maxL = tk.Label(f,text="Maximum")
        self.maxL.grid(row=0,column=5)
        self.nameE = []
        self.valueE = []
        self.fixedC = []
        self.stepE = []
        self.minE = []
        self.maxE = []
        for row in range(1,9):
            p = tk.Entry(f,width=9)
            p.grid(row=row,column=0)
            self.nameE.append(p)
            i = tk.Entry(f,width=9)
            i.grid(row=row,column=1)
            self.valueE.append(i)
            i = tk.IntVar()
            c = tk.Checkbutton(f,variable=i)
            c.grid(row=row,column=2)
            self.fixedC.append(i)
            step = tk.Entry(f,width=9)
            step.grid(row=row,column=3)
            self.stepE.append(step)
            min = tk.Entry(f,width=9)
            min.grid(row=row,column=4)
            self.minE.append(min)
            max = tk.Entry(f,width=9)
            max.grid(row=row,column=5)
            self.maxE.append(max)
        f.pack()
        tk.Button(self,text="Do the Fit",command=self.doFit,default="active").pack()
        self.result = tk.Label(self,text="")
        self.result.pack(fill=tk.X)
        tk.Label(self,text="\n(Re-drawing the plot will erase the fit.)").pack()
        self.methodChanged()
    def windowClosed(self):
        self.destroy()
        delattr(self.plot,"fitPanel")
    def hide(self):
        self.withdraw()
    def show(self):
        self.update()
        self.deiconify()
    def methodChanged(self,event=None):
        method = self.methodE.get()
        # these methods use limits (called bounds):
        if method in [ 'Powell', 'L-BFGS-B', 'TNC' ]:
            for e in self.minE+self.maxE: e.config(state='normal')
            self.minL.config(state='normal')
            self.maxL.config(state='normal')
            self.useBounds = True
        else:
            for e in self.minE+self.maxE: e.config(state='disabled')
            self.minL.config(state='disabled')
            self.maxL.config(state='disabled')
            self.useBounds = False
        # these methods use the initial step (called eps):
        if method in [ 'CG', 'BFGS', 'L-BFGS-B', 'SLSQP' ]:
            for e in self.stepE: e.config(state='normal')
            self.stepL.config(state='normal')
            self.useStep = True
        else:
            for e in self.stepE: e.config(state='disabled')
            self.stepL.config(state='disabled')
            self.useStep = False
    def setFit(self,fit):
        self.fit = fit
        self.setText(self.fExprE,fit.expression)
        self.methodE.set(self.fit.method)
        self.methodChanged()
        for i,p in enumerate(fit.params):
            self.setText(self.nameE[i],p.name)
            self.setText(self.valueE[i],p.value)
            self.fixedC[i].set(int(p.fixed))
            self.setText(self.stepE[i],p.step)
            self.setText(self.minE[i],p.min)
            self.setText(self.maxE[i],p.max)
    def setText(self,entry,value):
        entry.delete(0,9999)
        if value is not None: entry.insert(0,str(value))
    def putEntriesIntoFit(self):
        if self.fit is None:
            self.fit = Fit(self.plot,"")
        self.fit.params = []
        self.fit.expression = self.fExprE.get()
        self.fit.method = self.methodE.get()
        for i in range(len(self.nameE)):
            name = self.nameE[i].get()
            if name == "": continue
            try: value = float(self.valueE[i].get())
            except: value = 0.0
            fixed = (self.fixedC[i].get() != 0)
            try: step = float(self.stepE[i].get())
            except: step = None
            try: min = float(self.minE[i].get())
            except: min = None
            try: max = float(self.maxE[i].get())
            except: max = None
            self.fit.param(name,value,fixed,step,min,max)
        self.result.config(text="")
    def doFit(self):
        self.putEntriesIntoFit()
        ret = self.fit.doFit()
        values = self.fit.getValues()
        # present the results
        chisq = values["Chisq"]
        ndf = values["NDF"]
        iterations = values["Iterations"]
        message = values["Message"]
        if ret:
            self.result.config(text="Chi-sq=%g  NDF=%d  iterations=%d\n%s" % \
                                    (chisq,ndf,iterations,message))
        else:
            self.result.config(text= \
                        "FAILURE  Chi-sq=%g  NDF=%d  iterations=%d\n%s" % \
                                    (chisq,ndf,iterations,message))
        for i in range(len(self.nameE)):
            n = self.nameE[i].get()
            if n == "": continue
            v = values[n]
            self.setText(self.valueE[i],v)
    def plotFit(self):
        self.putEntriesIntoFit()
        self.fit.drawParams()

###
### class Fit -- do a fit to a plot
###     param() function adds a parameter
###     doFit() performs the fit
###         returns True if the fit converged, False if not
###     getValues() returns a dict with entries for "Chisq","NDF","Iterations",
###         "Message", and the current value of each parameter
###     chisq() is the function to be minimized
###     draw() will draw the result in self.plot
###
class Fit:
    # class Param is used to split Params into working and fixed
    class Param:
        def __init__(self,name,value,fixed,step,min,max):
            self.name = name
            self.value = value
            self.fixed = fixed
            self.step = step
            self.min = min
            self.max = max
    def __init__(self, plot, expression, method="Nelder-Mead",
                                                        printDetails=False):
        self.plot = plot
        self.plot.fit = self
        self.expression = expression
        self.method = method
        self.printDetails = printDetails
        self.params = []    # contains Param-s for all params
        # these temporaries are used only during the call to minimize():
        # (in self as they are used in both doFit() and chisq())
        self.names = None   # just non-fixed params
        self.values = None  # just non-fixed params
        self.steps = None   # just non-fixed params
        self.bounds = None  # just non-fixed params
        self.indexes = None # just non-fixed params
        self.fixed = None   # dict of fixed params
        # results of the fit:
        self.ret = None
        self.chisq_value = 0.0
        self.ndf = 0
        self.iterations = 0
        self.message = ""
        self._getDataFromPlot()
    def _getDataFromPlot(self):
        self.x,self.xExpr,self.y,self.yExpr,self.yerr,self.yerrExpr = \
                                                self.plot.fitValues()
        nDelete = 0
        # remove points with yerr==0
        if self.yerr is not None:
            zeros = (self.yerr == 0.0)
            nDelete = np.sum(zeros)
            if nDelete > 0:
                self.x = self.x[~zeros]
                self.y = self.y[~zeros]
                self.yerr = self.yerr[~zeros]
    def param(self, name, value, fixed=False, step=None, min=None, max=None):
        self.params.append(self.Param(name,value,fixed,step,min,max))
    def doFit(self):
        # split params into (names,values,steps,bounds,indexes) and fixed
        self.names = []
        self.values = []
        self.fixed = {}
        self.steps = []
        self.bounds = []
        self.indexes = []
        for i,p in enumerate(self.params):
            if p.fixed:
                self.fixed[p.name] = p.value
            else:
                self.names.append(p.name)
                self.values.append(p.value)
                self.steps.append(p.step)
                self.bounds.append( (p.min,p.max) )
                self.indexes.append(i)
        # get the data from the plot and do the fit
        self._getDataFromPlot()
        try:
            if self.method in [ 'CG', 'BFGS', 'L-BFGS-B', 'SLSQP' ]:
                options = {"eps":self.steps}
            else:
                options = None
            self.ret = minimize(self.chisq, self.values, bounds=self.bounds,
                                        method=self.method, options=options)
        except Exception as e:
            messagebox.showerror(title="AllPlot Error",
                message="exception in minimize():\n"+str(e))
            return
        # present the results
        if self.printDetails: print(self.ret)
        self.chisq_value = self.chisq(self.ret.x)
        self.ndf = len(self.x) - len(self.ret.x)
        self.iterations = self.ret.nit
        self.message = self.ret.message
        if self.ret.success:
            if self.printDetails:
                print("Chi-sq=%g  NDF=%d  iterations=%d\n%s" % \
                      (self.chisq_value,self.ndf,self.iterations,self.message))
        else:
            if self.printDetails:
                print("FAILURE  Chi-sq=%g  NDF=%d  iterations=%d\n%s" % \
                      (self.chisq_value,self.ndf,self.iterations,self.message))
        # put the current values back into self.params
        for i,v in enumerate(self.ret.x):
            self.params[self.indexes[i]].value = v
        # clear the temp arrays used only for the call to minimize()
        self.names = None
        self.values = None
        self.fixed = None
        self.steps = None
        self.bounds = None
        self.indexes = None
        # put values into FitPanel
        if self.plot.fitPanel is None:
            self.plot.fitPanel = FitPanel(self.plot.window,self.plot)
        self.plot.fitPanel.setFit(self)
        # draw the result and return
        self.draw()
        return self.ret.success
    def getValues(self):
        vals = {"Chisq":self.chisq_value,
                "NDF":self.ndf,
                "Iterations":self.iterations,
                "Message":self.message}
        for p in self.params:
            vals[p.name] = p.value
        return vals
    def chisq(self, currentParamValues):
        var = {"x":self.x}
        for i,k in enumerate(self.names):
            var[k] = currentParamValues[i]
        var.update(self.fixed)
        self.fVals = eval(self.expression,plot.Functions,var)
        if self.yerr is None:
            self.fRet = np.sum((self.fVals-self.y)**2)
        else:
            self.fRet = np.sum(((self.fVals-self.y)/self.yerr)**2)
        if self.printDetails:
            print("Chi-sq=%g " % self.fRet,end="")
            for i,k in enumerate(self.names):
                print("%s=%g " % (k,currentParamValues[i]),end="")
            print(" ")
        return self.fRet
    def drawParams(self):
        var = {"x":self.x}
        for p in self.params:
            if p.name is None or p.name == "": continue
            var[p.name] = p.value
        self.fVals = eval(self.expression,plot.Functions,var)
        self.draw()
    def draw(self):
        self.plot.draw()  # need to re-draw the plot first
        self.plot.axes.plot(self.x,self.fVals,"r-")
    def toAPIString(self):
        ret = ""
        ret += "fit = ap.Fit(p%d,'%s','%s',%d)\n" % \
           (self.plot.number,self.expression,self.method,int(self.printDetails))
        for p in self.params:
            ret += "fit.param('%s',%g,%d," % (p.name,p.value,p.fixed) + \
                    str(p.step)+","+str(p.min)+","+str(p.max)+")\n"
        ret += "fit.doFit()\n"
        return ret

###
### displayFitPanel()
###
def displayFitPanel(window):
    plot = window.currentPlot()
    if plot is None: return
    if plot.fitPanel is not None:
        plot.fitPanel.show()
    elif plot.fitValues()[0] is not None:
        FitPanel(window,plot)
    else:
        window.bell()

###
### generateDeltaTime()
###
def generateDeltaTime(window):
    try:
        data = window.input.getAllData()
        time = data["Time"]
    except:
        messagebox.showerror(title="AllPlot Error",
                    message="No 'Time' column.")
        return
    dt = time[1:] - time[:-1]
    # duplicate last entry so DeltaTime has same # entries as all the others
    dt = np.append(dt,(time[-1] - time[-2]))
    data["DeltaTime"] = dt
    msg = "DeltaTime created."
    if time.dtype == np.float32:
        msg += "\nData sre single precision with only 6 digits of precision, so these differences may be inaccurate."
    else:
        msg += "\nData are double precision, so these differences should be reasonably accurate."
    messagebox.showinfo(title="AllPlot", message=msg)


###
### class CreateColumnDialog
###
class CreateColumnDialog(simpledialog.Dialog):
    def __init__(self,window):
        self.window = window
        simpledialog.Dialog.__init__(self,window,"Create Column")
    def body(self,master):
        tk.Label(master,text="Create a new column using an expression that uses the\nfields (columns) of the data, and standard functions.").grid(row=0,column=0,columnspan=4)
        tk.Label(master,text="Column Name:").grid(row=1,column=0,sticky="E")
        self.name = tk.Entry(master)
        self.name.insert(tk.END,"NewColumn")
        self.name.grid(row=1,column=1,columnspan=3,sticky="EW")
        tk.Label(master,text="Enter expression using the column names:").grid(row=2,column=0,columnspan=4)
        columnNames = self.window.input.columnNames()
        tk.Label(master,text=",".join(columnNames)).grid(row=3,column=0,columnspan=4)
        tk.Label(master,text="Expression:").grid(row=4,column=0,sticky="E")
        self.expr = tk.Entry(master)
        self.expr.grid(row=4,column=1,columnspan=3,sticky="EW")
        self.deltas = tk.IntVar()
        tk.Checkbutton(master,text="Calculate deltas between successive entries",variable=self.deltas).grid(row=5,column=0,columnspan=4,sticky="W")
        self.movingAvg = tk.IntVar()
        tk.Checkbutton(master,text="Moving Average",variable=self.movingAvg).grid(row=6,column=0,sticky="W")
        tk.Label(master,text="  Nsamples:").grid(row=6,column=2,sticky="E")
        self.Nsamples = tk.Entry(master,width=6)
        self.Nsamples.grid(row=6,column=3,sticky="W")
    def validate(self):
        # get values from Entry-s
        try:
            name = self.name.get()
            expr = self.expr.get()
            Nsamples = 0
            deltas = self.deltas.get()
            movingAvg = self.movingAvg.get()
        except:
            messagebox.showerror(title="AllPlot Error",
                            message="Invalid entry.")
            return False
        try:
            if movingAvg:
                Nsamples = int(self.Nsamples.get())
                if Nsamples < 3:
                    messagebox.showerror(title="AllPlot Error",
                            message="Nsamples must be >= 3.")
                    return False
        except:
            messagebox.showerror(title="AllPlot Error",
                            message="Invalid value for Nsamples.")
            return False
        # check that name is unique
        columnNames = self.window.input.columnNames()
        if name in columnNames:
            messagebox.showerror(title="AllPlot Error",
                            message="Column '%s' already exists." % name)
            return False
        # evaluate expr (while checking for error)
        try:
            data = self.window.input.data
            self.newColumn = eval(expr,plot.Functions,data)
        except:
            messagebox.showerror(title="AllPlot Error",
                            message="Invalid expression.")
            return False
        try:
            if self.deltas.get():
                self.newColumn[1:] = self.newColumn[1:] - self.newColumn[0:-1]
                self.newColumn[0] = self.newColumn[1]
            if self.movingAvg.get():
                N = int(self.Nsamples.get())
                self.newColumn = np.convolve(self.newColumn,np.ones(N)/float(N),"same")
                self.newColumn[0:N//2] = self.newColumn[N//2]
                self.newColumn[-N//2+1:] = self.newColumn[-N//2]
        except:
            messagebox.showerror(title="AllPlot Error",
                            message="Internal Computation error.")
            return False
        return True
    def apply(self):
        self.window.input.data[self.name.get()] = self.newColumn

###
### createColumn()
###
def createColumn(window):
    CreateColumnDialog(window)

###
### histogramSampleIntervals()
###
def histogramSampleIntervals(window):
    class SampleHistogram(plot.Plot):
        def __init__(self, window):
            plot.Plot.__init__(self,window,"Intervals",forceAxes=None)
            self.window = window
            self.style = plot.Style()
            self.axesparam = plot.AxesParam()
            self.type = "IntervalHistogram"
            try:
                data = window.input.getAllData()
                time = data["Time"]
            except:
                messagebox.showerror(title="AllPlot Error",
                            message="No 'Time' column.")
                return
            self.draw()
        def draw(self):
            self.setupFrame()
            axes = self.axes
            time = self.window.input.getAllData()["Time"]
            x = time[1:] - time[:-1]
            y,edges = np.histogram(x, bins=100)
            centers = (edges[1:] + edges[:-1]) / 2.0
            self.xArray = centers
            self.yArray = y
            self.yerrArray = None
            drawstyle = "steps-mid"
            self.axesparam.apply(axes,edges,y)
            axes.errorbar(centers,y, 
                drawstyle=drawstyle, color=self.style.color,
                linestyle=self.style.linestyle, marker=self.style.pointMarker)
            mean = np.mean(x)
            sigma = np.std(x)
            text = r'$N=%d, \mu=%.6g, \sigma=%.6g$' % (len(x), mean, sigma)
            axes.set_title(text,loc="right")
            axes.set_xlabel("Sample Interval")
            axes.ticklabel_format(useOffset=False, style='plain')
    try:
        time = window.input.getAllData()["Time"]
    except:
        messagebox.showerror(title="AllPlot Error",
                    message="No 'Time' column to sample.")
        return
    SampleHistogram(window)


###
### class MarkerPanel -- control the creation of time markers; it is  a
### non-modal dialog that places itself to the right of the current plot
### window.
###
class MarkerPanel(tk.Toplevel):
    def __init__(self,window,plot):
        tk.Toplevel.__init__(self,window)
        self.window = window
        self.plot = plot
        self.plot.markerPanel = self
        self.protocol("WM_DELETE_WINDOW", self.windowClosed)
        xpos = window.winfo_x() + window.winfo_width()
        ypos = window.winfo_y()
        self.geometry("+%d+%d" % (xpos,ypos))
        self.title("Marker Panel "+self.plot.name)
        tk.Label(self,text="""
This marker panel will permit you to insert time markers into
the file. The current plot must have Time as its X axis;
<ALT>-Click will insert a marker, as will editing the text box.
The markers are displayed in every plot with Time as its X axis.
The markers are written when the file is saved (as .npz).""").pack()
        f = tk.Frame(self, bd=3, relief=tk.SUNKEN)
        self.text = scrolledtext.ScrolledText(f,height=16, width=20)
        self.text.pack()
        f.pack()
        data = window.input.getAllData()
        if "_markers" in data:
            markers = data["_markers"]
            for v in markers:
                self.text.insert(tk.END,"%.3f\n"%v)
        tk.Button(self,text="Update Markers",command=self.updateMarkers).pack()
        plot.register_click_callback(self.mouse_click)
    def windowClosed(self):
        self.destroy()
        delattr(self.plot,"markerPanel")
    def hide(self):
        self.withdraw()
    def show(self):
        self.update()
        self.deiconify()
    def mouse_click(self,event):
        if event.key == "alt" and event.name == "button_press_event":
            self.addMarker(event.xdata)
    def addMarker(self,time):
        self.text.insert(tk.END,str(round(time,3))+'\n')
        self.updateMarkers()
    def updateMarkers(self):
        try:
            m = self.text.get('1.0',tk.END).split()
            markers = []
            for i in m:
                if i == "": continue
                markers.append(round(float(i),3))
            data = self.window.input.getAllData()
            data["_markers"] = np.sort(np.array(markers))
            try: self.plot.draw()
            except: pass
        except:
            messagebox.showerror(title="AllPlot Error",
                message="Invalid entry")

###
### displayMarkerPanel()
###
def displayMarkerPanel(window):
    plot = window.currentPlot()
    if plot is None: return
    if plot.markerPanel is not None:
        plot.markerPanel.show()
    else:
        # use fitValues() to determine the X axis
        x,xExpr,y,yExpr,yerr,yerrExpr = plot.fitValues()
        if xExpr is not None and xExpr == "Time":
            MarkerPanel(window,plot)
        else:
            messagebox.showerror(title="AllPlot Error",
                message="Markers require the plot to have Time as its X-axis")

###
### editParameters()
###
def editParameters(window):
    inp = window.input
    print("Parameters:",inp.saveAllParams())

