#   window.py - Window for AllPlot

import sys
import os.path as path
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import simpledialog
from tkinter import scrolledtext
import tkinter.filedialog as filedialog
import zipfile

from input import DummyInput, Generate
import textinput, numpyinput, plot, tools
from help import helpAllPlot, helpAllPlotAPI

# MacOS has a Python/Tkinter bug that often disables menus; refreshing menus
# brings them back; this program does that whenever the Window menu changes 
# or whenever any window gets focus. That makes it usually unnoticeable, but 
# if necessary you can type an 'm' into any window to refresh them manually.
# Or you can switch to any other window, and then switch back again.

if __name__ == '__main__':  # test required by multiprocessing in iit.py
    print("TODO: enhance NumPy file to include parameters, markers, and comments")
    print("TODO: variables (saved expressions)")

###
### check if Root is installed
###
try:
    import root  # this is the local root.py, not pyroot
    RootInput = True
    print("Root input is available")
except:
    RootInput = False

###
### check if ADC is connected
###
try:
    import iit
    AdcInput = iit.ADC_detect()
except:
    AdcInput = False

###
### Globals
###
Title = "AllPlot"
RootWindow = None
DisplayAllBusy = False
WindowList = []
OpenInputDialog = None

###
### setup() and displayAll(), called by both the API and main(), so they
### must be safe to call multiple times. displayAll(*0 does not return until
### all windows have been closed.
###
def setup():
    global RootWindow
    if RootWindow is not None:
        RootWindow.withdraw()
        return
    RootWindow = tk.Tk()
    RootWindow.withdraw()
    RootWindow.protocol("WM_DELETE_WINDOW", quit)

def displayAll():
    global RootWindow, DisplayAllBusy
    if DisplayAllBusy:
        RootWindow.withdraw()
        return
    DisplayAllBusy = True
    RootWindow.mainloop()
    DisplayAllBusy = False

###
### quit() - destroy all windows, which exits the app
###
def quit():
    global WindowList, RootWindow
    for w in WindowList:
        w.input.windowClosed()
    try: RootWindow.destroy() # may already have been destroyed
    except: pass

###
### fileOpen() - function for the file/open menu item
### This displays a dialog box from the input module.
### The program starts up by calling fileOpen().
###
def fileOpen():
    global WindowList, RootWindow
    input = openInput(RootWindow)
    if input is None: 
        # in case the initial open was canceled
        if len(WindowList) == 0:
            try: RootWindow.destroy()
            except: pass
        return
    return Window(input)

###
### class Window -- implements all windows (but not dialog boxes)
###
### Note plotList[] can contain None (e.g. for a Generate tab with no plot).
###
class Window(tk.Toplevel):
    def __init__(self,input):
        global Title, WindowList
        print("AllPlotWindow filename: '"+input.shortFilename+"'   nData:",
                                                                input.nData())
        tk.Toplevel.__init__(self, takefocus=True) # super() not portable 2/3
        self.title(Title+"  "+input.shortFilename)
        WindowList.append(self)
        self.input = input
        if not input.hasData(): input.read()
        self.plotList = []
        self.nextPlotNumber = 1
        self.autoReloadInterval = 0     # # seconds between reloads, 0=none
        self.setupMenu()
        self.geometry("%dx%d" % (1024,760))
        self.topFrame = tk.Frame(self)
        self.topFrame.pack(fill=tk.X)
        self.ypos = tk.Label(self.topFrame,text="",width=15, anchor=tk.NW)
        self.ypos.pack(side=tk.RIGHT)
        self.xpos = tk.Label(self.topFrame,text="",width=15, anchor=tk.NW)
        self.xpos.pack(side=tk.RIGHT)
        self.selection = tk.Label(self.topFrame,text="")
        self.selection.pack(side=tk.RIGHT)
        self.tabs = ttk.Notebook(self)
        self.tabs.pack(fill=tk.BOTH,expand=True)
        self.tabs.bind("<<NotebookTabChanged>>",self.tabChanged)
        if sys.platform == "darwin":
            self.tabs.bind("<Button-2>",self.tabContextMenu)
        else:
            self.tabs.bind("<Button-3>",self.tabContextMenu)
        self.tabs.enable_traversal()
        self.input.setWindow(self)
        # almost every window gets a Selection Plot coupled to its input
        tmp = self.input.getSelectionPlotColumnName()
        if tmp is not None and tmp != "(none)":
            plot.SelectionPlot(self)
        # Input.windowClosed() must call window.windowClosed()
        self.protocol("WM_DELETE_WINDOW", self.input.windowClosed)
        # these events are used to refresh the menu, workaround for Mac OS bug
        self.bind("<Key>",self.keyEvent)
        self.bind("<FocusIn>",self.focusEvent)
    def windowClosed(self):
        global RootWindow, WindowList
        if self in WindowList: WindowList.remove(self)
        self.setupMenu() # update the window list
        if len(WindowList) == 0:
           RootWindow.destroy() # also destroys any remaining windows; app exits
        else:
           self.destroy()
    # setupMenu is called by focusEvent() and by setActive() -- that
    # which updates the menus, including the window list.
    def setupMenu(self):
        global WindowList, RootWindow
        self.menubar = tk.Menu(self)
        self.config(menu=self.menubar)
        filemenu = tk.Menu(self.menubar)
        self.menubar.add_cascade(label="File",menu=filemenu)
        filemenu.add_command(label="Open",command=fileOpen,accelerator="Ctrl-O")
        RootWindow.bind_all("<Control-o>", lambda event: fileOpen())
        filemenu.add_separator()
        filemenu.add_command(label="Reload",command=self.reload,accelerator="Ctrl-R")
        RootWindow.bind_all("<Control-r>", lambda event: self.reload())
        filemenu.add_command(label="Reload On Change",command=self.reloadOnChange)
        filemenu.add_command(label="Reload 30 Sec.",command=self.reload30Seconds)
        filemenu.add_separator()
        filemenu.add_command(label="Save Data (.npz)",command=self.saveData)
        filemenu.add_command(label="Save Image",command=self.saveImage)
        filemenu.add_command(label="Save Config",command=self.saveConfig)
        filemenu.add_command(label="Load Config",command=self.loadConfig,accelerator="Ctrl-L")
        RootWindow.bind_all("<Control-l>", lambda event: self.loadConfig())
        filemenu.add_separator()
        filemenu.add_command(label="Generate",command=self.generate,accelerator="Ctrl-G")
        RootWindow.bind_all("<Control-g>", lambda event: self.generate())
        filemenu.add_separator()
        filemenu.add_command(label="Quit",command=quit)
        plot.setupPlotMenu(self)
        tools.setupToolMenu(self)
        windowMenu = tk.Menu(self.menubar)
        self.menubar.add_cascade(label="Window",menu=windowMenu)
        for w in WindowList:
            name = w.input.shortFilename
            windowMenu.add_command(label=name, command=w.setActive)
        setHelpMenu(self.menubar)
    def reload(self):  # re-read the file, re-draw all the plots
        self.autoReloadInterval = 0
        self.input.reload()
        self.redrawAll()
    def reloadOnChange(self):
        self.autoReloadInterval = 1
        self.autoReload()
    def reload30Seconds(self):
        self.autoReloadInterval = 30
        self.autoReload()
    def autoReload(self):
        if self.input.isChanged():
            self.input.reload()
            self.redrawAll()
        if self.autoReloadInterval > 0:
            self.after(self.autoReloadInterval,self.autoReload)
    def redrawAll(self):
        for p in self.plotList:
            if p is not None: p.reload()
    def saveData(self):
        fn = filedialog.asksaveasfilename(parent=self,title="Save Data")
        if fn is not None:
            if not fn.endswith(".npz"): fn += ".npz"
            data = self.input.getAllData()
            np.savez_compressed(fn,**data)
    def saveImage(self):
        i = self.tabs.index("current")
        if self.plotList[i] is None: return
        figure = self.plotList[i].figure
        fn = filedialog.asksaveasfilename(parent=self,title="Save Image")
        if fn is not None:
            if fn.endswith(".png") or fn.endswith(".svg") or \
               fn.endswith(".jpg") or fn.endswith(".jpeg"):
                pass
            else:
                fn += ".png"
            figure.savefig(fn)
    def saveConfig(self):
        fn = filedialog.asksaveasfilename(parent=self,title="Save Config")
        if fn is not None and fn != '':
            try:
                with open(fn,"w") as f:
                    f.write("# 'import AllPlotAPI as ap' already done.\n")
                    for w in WindowList:
                        f.write(w.toAPIString())
                    f.write("\nap.displayAll()\n")
            except:
                messagebox.showerror(Title+" Error","Cannot write '%s'" % fn)
    def loadConfig(self):
        fn = filedialog.askopenfilename(parent=self,title="Load Config")
        if fn is None or fn == '': return
        try:
            with open(fn) as f:
                text = f.read()
        except:
            messagebox.showerror(Title+" Error","Cannot read '%s'" % fn)
            return
        try:
            import AllPlotAPI as ap
            print("Loading config '%s'" % fn)
            exec(text,{"ap":ap},{})
        except BaseException as e:
            messagebox.showerror(Title+" Error",
                    "Exception in '%s': %s" % (fn,str(e)))
    def generate(self):
        Window(Generate())
    def keyEvent(self,event):
        if event.char == "m" or event.char == "M":
            print("Manual Refresh MenuBar")
            self.setupMenu()
    def focusEvent(self,event):
        # MacOS has a bug that often disables menus; refreshing them fixes it
        self.setupMenu()
    def setActive(self):
        self.lift()
        self.focus_force()
        self.grab_set()
        self.grab_release()
        self.setupMenu()
    def tabChanged(self,event):
        # show the FitPanel and MarkerPanel (if any) for the now-current plot,
        # hiding any others
        c = self.currentPlot()
        for p in self.plotList:
            if p is None: continue
            if p is c:
                if p.fitPanel is not None: p.fitPanel.show()
                if p.markerPanel is not None: p.markerPanel.show()
            else:
                if p.fitPanel is not None: p.fitPanel.hide()
                if p.markerPanel is not None: p.markerPanel.hide()
    def tabContextMenu(self,event):
        if event.widget.identify(event.x, event.y) == 'label':
            tabname = "@%d,%d" % (event.x, event.y)
            event.widget.select(tabname)
            m = tk.Menu(event.widget, tearoff=0)
            m.add_command(label="Replot", command=lambda: \
                                            self.currentPlot().draw())
            m.add_command(label="Edit Plot Style", command=lambda: \
                                            plot.editStyle(self))
            m.add_separator()
            m.add_command(label="Delete Plot", command=lambda: \
                                            plot.deletePlot(self))
            m.grab_set()
            m.tk_popup(event.x_root,event.y_root)
            m.grab_release()
    def currentPlot(self):
        i = self.tabs.index("current")
        if i < 0 or i >= len(self.plotList): return None
        return self.plotList[i]
    def plotNumber(self):
        self.nextPlotNumber += 1
        return self.nextPlotNumber - 1
    ###
    ### listAllPlots() - return a list of tuples (filename,plotname,Plot) --
    ### the two names are strings, and the third entry is a reference to the
    ### Plot object.
    ###
    def listAllPlots(self):
        global WindowList
        ret = []
        for w in WindowList:
            for p in w.plotList:
                if p is None: continue
                ret.append( (w.input.filename,p.getName(),p) )
        return ret
    ###
    ### create an extra window for tools to display results
    ### the extra window starts with the same size as normal windows
    ###
    def createExtraWindow(self):
        win = Window(DummyInput())
        # make it be initially be empty
        win.topFrame.destroy()
        win.tabs.destroy()
        return win
    ###
    ### create a non-modal dialog window
    ### the new window will size itself to its content
    ###
    def createNonModalDialog(self):
        win = tk.TopLevel(self)
        return win
    ###
    ### toAPIString() - return a string suitable for use with the API.
    ### Every toAPIString() assumes "import AllPlot as ap", and ends with '\n'.
    ### Window.toAPIString() includes its input, window, and each plot.
    ### Each plot.toAPIString() includes its Style and AxesParam.
    ### Local variables: the Input is in inp, the Window is in win,
    ### Style in style, AxesParam ### in axesparam, and each Plot is in p#
    ### where # is its number.
    ###
    def toAPIString(self):
        ret = "\n"
        ret += self.input.toAPIString()
        ret += "win = ap.Window(inp)\n"
        for p in self.plotList:
            if p is None: continue
            ret += p.toAPIString()
        ret += "\n"
        return ret


###
### openInput() - display a dialog box to select type of input, a file of
### that type, then open it.
###
def openInput(master):
    class PopUpWindow(tk.Toplevel):
        def __init__(self,msg):
            tk.Toplevel.__init__(self)
            self.title(Title)
            self.geometry("400x150")
            tk.Label(self,text=msg,wraplength=300). \
                                            pack(fill=tk.BOTH,expand=True)
            self.update()
    class InputDialog(simpledialog.Dialog):
        def __init__(self, master):
            global OpenInputDialog
            OpenInputDialog = self
            self.result = None
            simpledialog.Dialog.__init__(self,master,title=Title+" Input")
            OpenInputDialog = None
        def body(self,master):
            tk.Label(master,text="""
A data file consists of one or more columns (multiple values),
zero or more parameters (single values), plus an optional 
comment. Plots display columns or expressions as a two-
dimensional graph. Select the input-file format:""").pack()
            self.tabs = ttk.Notebook(master)
            self.tabs.pack()
            self.param = []
            # text file
            p = textinput.TextInputParameters()
            if p.isValid():
                self.param.append(p)
                p.addTab(self.tabs)
            # NumPy File
            p = numpyinput.NumpyInputParameters()
            if p.isValid():
                self.param.append(p)
                p.addTab(self.tabs)
            # Root File
            if RootInput:
                p = root.RootInputParameters()
                if p.isValid():
                    self.param.append(p)
                    p.addTab(self.tabs)
            # ADC Box
            if AdcInput:
                p = iit.AdcInputParameters()
                if p.isValid():
                    self.param.append(p)
                    p.addTab(self.tabs)
        def apply(self):
            self.result = self.param[self.tabs.index('current')].getInput()
    d = InputDialog(master)
    if d.result is not None:
        d.result.popup = PopUpWindow("Reading %s" % d.result.filename)
        d.result.read()
    return d.result

def rootGenerate(event=None):
    global RootWindow, OpenInputDialog
    if OpenInputDialog is not None: 
        try: OpenInputDialog.destroy()
        except: pass
        OpenInputDialog = None
    Window(Generate())

def rootLoadConfig(event=None):
    global RootWindow, OpenInputDialog
    fn = filedialog.askopenfilename(parent=RootWindow,title="Load Config")
    if fn is None or fn == "": 
        print("rootLoadConfig no file")
        return
    try:
        with open(fn) as f:
            text = f.read()
    except:
        messagebox.showerror(Title+" Error","Cannot read '%s'" % fn)
        print("rootLoadConfig error")
        return
    if OpenInputDialog is not None: 
        try: OpenInputDialog.destroy()
        except: pass
        OpenInputDialog = None
    try:
        import AllPlotAPI as ap
        print("Loading config '%s'" % fn)
        exec(text,{"ap":ap},{})
    except BaseException as e:
        messagebox.showerror(Title+" Error",
                "Exception in '%s': %s" % (fn,str(e)))
        RootWindow.destroy()
    RootWindow.deiconify()
    RootWindow.withdraw()

###
### Help Menu and functions
###
def setHelpMenu(menubar):
    helpmenu = tk.Menu(menubar)
    menubar.add_cascade(label="Help",menu=helpmenu)
    helpmenu.add_command(label="About "+Title, command=about)
    helpmenu.add_separator()
    helpmenu.add_command(label=Title+" Help", command=help)
    helpmenu.add_command(label="AllPlot API Help", command=helpAPI)
def about():
    msg = ''' AllPlot 3.0
Copyright (C) 2020-23 by Tom Roberts.
All rights reserved.

This program is Open Source and may be
redistributed in accordance with the Gnu
Public License v3.'''
    messagebox.showinfo(title=Title, message=msg)
def help():
    global RootWindow
    helpText = helpAllPlot()
    w = tk.Toplevel(RootWindow)
    w.title(Title+" Help")
    text = scrolledtext.ScrolledText(w,padx=5)
    text.insert(tk.END,helpText)
    text.config(state=tk.DISABLED)
    text.pack(expand=True, fill='both')
    # need this for the initial input dialog:
    text.grab_set()
    text.grab_release()
def helpAPI():
    global RootWindow
    helpText = helpAllPlotAPI()
    w = tk.Toplevel(RootWindow)
    w.title("AllPlot API Help")
    text = scrolledtext.ScrolledText(w,padx=5)
    text.insert(tk.END,helpText)
    text.config(state=tk.DISABLED)
    text.pack(expand=True, fill='both')
    # need this for the initial input dialog:
    text.grab_set()
    text.grab_release()

###
### Application starts up by displaying the fileOpen dialog box
###
def main():
    global RootWindow
    # We don't use the RootWindow, but need it for the fileOpen dialog.
    # It gets destroyed in windowClosed() when the last Window is closed.
    setup()
    # initial menu
    menubar = tk.Menu(RootWindow)
    RootWindow.config(menu=menubar)
    filemenu = tk.Menu(menubar)
    menubar.add_cascade(label="File",menu=filemenu)
    filemenu.add_command(label="Generate",command=rootGenerate,accelerator="Ctrl-G")
    RootWindow.bind_all("<Control-g>", rootGenerate)
    # rootLoadConfig() does indeed load the configuration and display its
    # windows and plots. But it won't withdraw RootWindow, and the mouse
    # does not work in the windows from the config. Window.loadConfig() works.
    filemenu.add_command(label="Load Config",command=rootLoadConfig,accelerator="Ctrl-L")
    RootWindow.bind_all("<Control-l>", rootLoadConfig)
    setHelpMenu(menubar)
    # present the fileOpen dialog box to select an input and its window
    w = fileOpen()
    displayAll()
    # make the ADC flush its output and close the ADC device
    if AdcInput: iit.terminate()
    print(Title+" is exiting")

